class SearchMethods(object):
    """Search methods of the public API."""

    def search(self, search_term, per_page=None, page=None, type_=""):
        """Searches Genius.

        Args:
            search_term (:obj:`str`): A term to search on Genius.
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').
            type_ (:obj:`str`, optional): Type of item to search for
                ('song', 'lyric', 'artist', 'album', 'video',
                'article', 'user' or 'multi').

        Returns:
            :obj:`dict`

        .. Note::
            Specifying no :obj:`type_` parameter (which defaults to ``''``) or
            setting it as ``song`` will return the same results. Both will return
            songs. The only different is they return the hits in different
            keys:

                * ``type_=''``: ``response['hits']``
                * ``type_='song'``: ``response['sections'][0]['hits']``

            By Setting the type as ``multi`` the method will perform a search
            for all the other types and return an extra section called ``top hits``.

        .. Note::
            Instead of calling this method directly and specifying a type, you
            can use the alias methods.

        """
        if type_ == "":
            path = "search"
        else:
            path = "search/" + type_
        params = {"q": search_term, "per_page": per_page, "page": page}
        return self._make_request(path, params_=params, public_api=True)

    def search_albums(self, search_term, per_page=None, page=None):
        """Searches the albums on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "album"
        return self.search(search_term, per_page, page, endpoint)

    def search_articles(self, search_term, per_page=None, page=None):
        """Searches the articles on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "article"
        return self.search(search_term, per_page, page, endpoint)

    def search_artists(self, search_term, per_page=None, page=None):
        """Searches the artists on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "artist"
        return self.search(search_term, per_page, page, endpoint)

    def search_lyrics(self, search_term, per_page=None, page=None):
        """Searches the lyrics on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "lyric"
        return self.search(search_term, per_page, page, endpoint)

    def search_songs(self, search_term, per_page=None, page=None):
        """Searches the songs on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "song"
        return self.search(search_term, per_page, page, endpoint)

    def search_users(self, search_term, per_page=None, page=None):
        """Searches the users on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "user"
        return self.search(search_term, per_page, page, endpoint)

    def search_videos(self, search_term, per_page=None, page=None):
        """Searches the videos on Genius.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "video"
        return self.search(search_term, per_page, page, endpoint)

    def search_all(self, search_term, per_page=None, page=None):
        """Searches all types.

        Including: albums, articles, lyrics, songs, users and
        videos.

        Alias for :meth:`search() <PublicAPI.search>`

        Args:
            search_term (:obj:`str`): A term to search on Genius.
            per_page (:obj:`int`, optional): Number of results to
                return per page. It can't be more than 5 for this method.
            page (:obj:`int`, optional): Number of the page.

        Returns:
            :obj:`dict`

        Note:
            This method will also return a ``top hits`` section
            alongside other types.

        """
        endpoint = "multi"
        return self.search(search_term, per_page, page, endpoint)



from .album import AlbumMethods
from .annotation import AnnotationMethods
from .article import ArticleMethods
from .artist import ArtistMethods
from .cover_art import CoverArtMethods
from .discussion import DiscussionMethods
from .leaderboard import LeaderboardMethods
from .misc import MiscMethods
from .question import QuestionMethods
from .referent import ReferentMethods
from .search import SearchMethods
from .song import SongMethods
from .user import UserMethods
from .video import VideoMethods



class LeaderboardMethods(object):
    """Leaderboard methods of the public API."""

    def leaderboard(
        self, time_period="day", per_page=None, page=None, text_format=None
    ):
        """Gets the Genius community leaderboard.

        This method gets data of the community charts on the Genius.com page.

        Args:
            time_period (:obj:`str`, optional): Time period of the results.
                ('day', 'week', 'month' or 'all_time').
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        path = "leaderboard"
        params = {
            "time_period": time_period,
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=path, params_=params, public_api=True)

    def charts(
        self,
        time_period="day",
        chart_genre="all",
        per_page=None,
        page=None,
        text_format=None,
        type_="songs",
    ):
        """Gets the Genius charts.

        This method gets data of the chart on the Genius.com page.

        Args:
            time_period (:obj:`str`, optional): Time period of the results.
                The default is `all`.
                ('day', 'week', 'month' or 'all_time').
            chart_genre (:obj:`str`, optional): The genre of the results.
                The default value is ``all``.
                ('all', 'rap', 'pop', 'rb', 'rock' or 'country')
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').
            type_ (:obj:`int`, optional): The type to get the charts for.
                The default is ``songs``.
                ('songs', 'albums', 'artists' or 'referents').

        Returns:
            :obj:`dict`

        .. Note::
            The *referents* mentioned in the description of the :obj:`type_`
            argument is shown as *Lyrics* in the drop-down menu on Genius.com
            where you choose the *Type*.

        """
        endpoint = type_ + "/chart"
        params = {
            "time_period": time_period,
            "chart_genre": chart_genre,
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)



class UserMethods(object):
    """User methods of the public API."""

    def user(self, user_id, text_format=None):
        """Gets data for a specific user.

        Args:
            user_id (:obj:`int`): Genius user ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        path = "users/{}".format(user_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path, params_=params, public_api=True)

    def user_accomplishments(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
    ):
        """Gets user's accomplishments.

        This methods gets the section titled "TOP ACCOMPLISHMENTS" in
        the user's profile.

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).

        Returns:
            :obj:`dict`

        """
        endpoint = "users/{}/accomplishments".format(user_id)
        params = {"next_cursor": next_cursor, "per_page": per_page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def user_following(self, user_id, per_page=None, page=None):
        """Gets the accounts user follows.

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "users/{}/followed_users".format(user_id)
        params = {"page": page, "per_page": per_page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def user_followers(self, user_id, per_page=None, page=None):
        """Gets user's followers.

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "users/{}/followers".format(user_id)
        params = {"page": page, "per_page": per_page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def user_contributions(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
        sort=None,
        text_format=None,
        type_=None,
    ):
        """Gets user's contributions.

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').
            type_ (:obj:`int`, optional): Type of the contribution
                ('annotations', 'articles', 'pyongs', 'questions_and_answers',
                'comments', 'transcriptions' or 'unreviewed annotations').

        Returns:
            :obj:`dict`


        Note:
            Not all types support a sorting preference. Setting the :obj:`sort` for
            these types won't result in erros, but won't make a difference in the
            results either. To find out which types support which features, look at
            the alias methods.

        Note:
            Setting no value for the :obj:`type_` will return the user's contributions
            (regardless of its type) in chronological order; just like visting a
            user's profile page and scrolling down, looking at their contributions over
            time.

        """
        endpoint = "users/{}/contributions".format(user_id)
        if type_ is not None:
            endpoint += "/{}".format(type_)
        params = {
            "next_cursor": next_cursor,
            "per_page": per_page,
            "sort": sort,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def user_annotations(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
        sort="popularity",
        text_format=None,
    ):
        """Gets user's annotations.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            sort=sort,
            text_format=text_format,
            type_="annotations",
        )

    def user_articles(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
        sort="popularity",
        text_format=None,
    ):
        """Gets user's articles.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            sort=sort,
            text_format=text_format,
            type_="articles",
        )

    def user_pyongs(self, user_id, per_page=None, next_cursor=None, text_format=None):
        """Gets user's Pyongs.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            text_format=text_format,
            type_="pyongs",
        )

    def user_questions_and_answers(
        self, user_id, per_page=None, next_cursor=None, text_format=None
    ):
        """Gets user's Q&As.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            text_format=text_format,
            type_="questions_and_answers",
        )

    def user_suggestions(
        self, user_id, per_page=None, next_cursor=None, text_format=None
    ):
        """Gets user's suggestions (comments).

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            text_format=text_format,
            type_="comments",
        )

    def user_transcriptions(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
        sort="popularity",
        text_format=None,
    ):
        """Gets user's transcriptions.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            sort=sort,
            text_format=text_format,
            type_="transcriptions",
        )

    def user_unreviewed(
        self,
        user_id,
        per_page=None,
        next_cursor=None,
        sort="popularity",
        text_format=None,
    ):
        """Gets user's unreviewed annotations.

        Alias for :meth:`user_contributions() <PublicAPI.user_contributions>`

        This method gets user annotations that have the
        "This annotations is unreviewed" sign above them.

        Args:
            user_id (:obj:`int`): Genius user ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`str`, optional): Paginated offset
                (address of the next cursor).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.user_contributions(
            user_id=user_id,
            next_cursor=next_cursor,
            per_page=per_page,
            sort=sort,
            text_format=text_format,
            type_="unreviewed_annotations",
        )



from ..utils import convert_to_datetime
from .artist import Artist
from .base import BaseEntity


class Album(BaseEntity):
    """An album from the Genius.com database."""

    def __init__(self, client, json_dict, tracks):
        body = json_dict
        super().__init__(body["id"])
        self._body = body
        self._client = client
        self.artist = Artist(client, body["artist"])
        self.tracks = tracks
        self.release_date_components = convert_to_datetime(
            body.get("release_date_components")
        )

        self.api_path = body.get("api_path")
        self.cover_art_thumbnail_url = body.get("cover_art_thumbnail_url")
        self.cover_art_url = body.get("cover_art_url")
        self.full_title = body.get("full_title")
        self.name = body.get("name")
        self.name_with_artist = body.get("name_with_artist")
        self.url = body.get("url")

    def to_dict(self):
        body = super().to_dict()
        body["tracks"] = [track.to_dict() for track in self.tracks]
        return body

    def to_json(self, filename=None, sanitize=True, ensure_ascii=True):
        data = self.to_dict()

        return super().to_json(
            data=data, filename=filename, sanitize=sanitize, ensure_ascii=ensure_ascii
        )

    def to_text(self, filename=None, sanitize=True):
        data = "\n\n".join(
            f"[Song {n}: {track.song.title}]\n{track.song.lyrics}"
            for n, track in enumerate(self.tracks, start=1)
        ).strip()

        return super().to_text(data=data, filename=filename, sanitize=sanitize)

    def save_lyrics(
        self,
        filename=None,
        extension="json",
        overwrite=False,
        ensure_ascii=True,
        sanitize=True,
        verbose=True,
    ):
        if filename is None:
            filename = "Lyrics_" + self.name.replace(" ", "")

        return super().save_lyrics(
            filename=filename,
            extension=extension,
            overwrite=overwrite,
            ensure_ascii=ensure_ascii,
            sanitize=sanitize,
            verbose=verbose,
        )


class Track(BaseEntity):
    """docstring for Track"""

    def __init__(self, client, json_dict, lyrics):
        from .song import Song

        body = json_dict
        super().__init__(body["song"]["id"])
        self._body = body
        self.song = Song(client, body["song"], lyrics)

        self.number = body["number"]

    def to_dict(self):
        body = super().to_dict()
        body["song"] = self.song.to_dict()
        return body

    def to_json(self, filename=None, sanitize=True, ensure_ascii=True):
        data = self.to_dict()

        return super().to_json(
            data=data, filename=filename, sanitize=sanitize, ensure_ascii=ensure_ascii
        )

    def to_text(self, filename=None, sanitize=True):
        data = self.song.lyrics

        return super().to_text(data=data, filename=filename, sanitize=sanitize)

    def save_lyrics(
        self,
        filename=None,
        extension="json",
        overwrite=False,
        ensure_ascii=True,
        sanitize=True,
        verbose=True,
    ):
        if filename is None:
            filename = "Lyrics_{:02d}_{}".format(self.number, self.song.title)
            filename = filename.replace(" ", "")

        return super().save_lyrics(
            filename=filename,
            extension=extension,
            overwrite=overwrite,
            ensure_ascii=ensure_ascii,
            sanitize=sanitize,
            verbose=verbose,
        )

    def __repr__(self):
        name = self.__class__.__name__
        return "{}(number, song)".format(name)



class SongMethods(object):
    """Song methods of the public API."""

    def song(self, song_id, text_format=None):
        """Gets data for a specific song.

        Args:
            song_id (:obj:`int`): Genius song ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "songs/{}".format(song_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def song_activity(self, song_id, per_page=None, page=None, text_format=None):
        """Gets activities on a song.

        Args:
            song_id (:obj:`int`): Genius song ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "songs/{}/activity_stream/line_items".format(song_id)
        params = {
            "text_format": text_format or self.response_format,
            "per_page": per_page,
            "page": page,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def song_comments(self, song_id, per_page=None, page=None, text_format=None):
        """Gets the comments on a song.

        Args:
            song_id (:obj:`int`): Genius song ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "songs/{}/comments".format(song_id)
        params = {
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def song_contributors(self, song_id):
        """Gets the contributors of a song.

        This method will return users who have contributed
        to this song by editing lyrics or song details.

        Args:
            song_id (:obj:`int`): Genius song ID

        Returns:
            :obj:`dict`

        """
        endpoint = "songs/{}/contributors".format(song_id)
        return self._make_request(path=endpoint, public_api=True)



class ReferentMethods(object):
    """Referent methods of the public API."""

    def referent(self, referent_ids, text_format=None):
        """Gets data of one or more referents.

        This method can get multiple referents in one call,
        thus increasing performance.

        Args:
            referent_ids (:obj:`list`): A list of referent IDs.
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Note:
            Using this method you can get the referent itself instead of
            the referents of a song or webpage which is what
            :meth:`referents() <PublicAPI.referents>` gets.

        """
        params = {"text_format": text_format or self.response_format}
        if len(referent_ids) == 1:
            endpoint = "referents/{}".format(referent_ids[0])
        else:
            endpoint = "referents/multi"
            params = [("text_format", params["text_format"])]
            for id in referent_ids:
                params.append(("ids[]", id))

        return self._make_request(path=endpoint, params_=params, public_api=True)

    def referents(
        self,
        song_id=None,
        web_page_id=None,
        created_by_id=None,
        per_page=None,
        page=None,
        text_format=None,
    ):
        """Gets item's referents

        You must supply :obj:`song_id`, :obj:`web_page_id`, or :obj:`created_by_id`.

        Args:
            song_id (:obj:`int`, optional): song ID
            web_page_id (:obj:`int`, optional): web page ID
            created_by_id (:obj:`int`, optional): User ID of the contributer
                who created the annotation(s).
            per_page (:obj:`int`, optional): Number of results to
                return per page. It can't be more than 50.
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        msg = "Must supply `song_id`, `web_page_id`, or `created_by_id`."
        assert any([song_id, web_page_id, created_by_id]), msg
        msg = "Pass only one of `song_id` and `web_page_id`, not both."
        assert bool(song_id) ^ bool(web_page_id), msg

        endpoint = "referents"
        params = {
            "song_id": song_id,
            "web_page_id": web_page_id,
            "created_by_id": created_by_id,
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(endpoint, params_=params, public_api=True)

    def referents_charts(
        self,
        time_period="day",
        chart_genre="all",
        per_page=None,
        page=None,
        text_format=None,
    ):
        """Gets the referents (lyrics) charts.

        Alias for :meth:`charts() <PublicAPI.charts>`.

        Args:
            time_period (:obj:`str`, optional): Time period of the results
                ('day', 'week', 'month' or 'all_time').
            chart_genre (:obj:`str`, optional): The genre of the results.
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.charts(
            time_period=time_period,
            chart_genre=chart_genre,
            per_page=per_page,
            page=page,
            text_format=text_format,
            type_="referents",
        )



import os

from lyricsgenius import Genius

# Import client access token from environment variable
access_token = os.environ.get("GENIUS_ACCESS_TOKEN", None)
assert access_token is not None, (
    "Must declare environment variable: GENIUS_ACCESS_TOKEN"
)

# Genius client
genius = Genius(access_token, sleep_time=1.0, timeout=15, retries=3)



# LyricsGenius
# copyright 2025 John W. R. Miller
# See LICENSE for details.

from filecmp import cmp

from .album import Album
from .artist import Artist
from .base import BaseEntity, Stats


class Song(BaseEntity):
    """A song from the Genius.com database."""

    def __init__(self, client, json_dict, lyrics=""):
        body = json_dict
        super().__init__(body["id"])
        self._body = body
        self._client = client
        self.artist = body["primary_artist"]["name"]
        self.lyrics = lyrics if lyrics else ""
        self.primary_artist = Artist(client, body["primary_artist"])
        self.stats = Stats(body["stats"])
        self.album = Album(client, body["album"], []) if body.get("album") else None

        self.annotation_count = body["annotation_count"]
        self.api_path = body["api_path"]
        self.full_title = body["full_title"]
        self.header_image_thumbnail_url = body["header_image_thumbnail_url"]
        self.header_image_url = body["header_image_url"]
        self.lyrics_owner_id = body["lyrics_owner_id"]
        self.lyrics_state = body["lyrics_state"]
        self.path = body["path"]
        self.pyongs_count = body["pyongs_count"]
        self.song_art_image_thumbnail_url = body["song_art_image_thumbnail_url"]
        self.song_art_image_url = body["song_art_image_url"]
        self.title = body["title"]
        self.title_with_featured = body["title_with_featured"]
        self.url = body["url"]

    def to_dict(self):
        body = super().to_dict()
        body["artist"] = self.artist
        body["lyrics"] = self.lyrics
        return body

    def to_json(self, filename=None, sanitize=True, ensure_ascii=True):
        data = self.to_dict()
        return super().to_json(
            data=data, filename=filename, sanitize=sanitize, ensure_ascii=ensure_ascii
        )

    def to_text(self, filename=None, sanitize=True):
        data = self.lyrics

        return super().to_text(data=data, filename=filename, sanitize=sanitize)

    def save_lyrics(
        self,
        filename=None,
        extension="json",
        overwrite=False,
        ensure_ascii=True,
        sanitize=True,
        verbose=True,
    ):
        if filename is None:
            filename = "Lyrics_{}_{}".format(
                self.artist.replace(" ", ""), self.title.replace(" ", "")
            ).lower()

        return super().save_lyrics(
            filename=filename,
            extension=extension,
            overwrite=overwrite,
            ensure_ascii=ensure_ascii,
            sanitize=sanitize,
            verbose=verbose,
        )

    def __str__(self):
        """Return a string representation of the Song object."""
        if len(self.lyrics) > 100:
            lyr = self.lyrics[:100] + "..."
        else:
            lyr = self.lyrics[:100]
        return '"{title}" by {artist}:\n    {lyrics}'.format(
            title=self.title, artist=self.artist, lyrics=lyr.replace("\n", "\n    ")
        )

    def __cmp__(self, other):
        return (
            cmp(self.title, other.title)
            and cmp(self.artist, other.artist)
            and cmp(self.lyrics, other.lyrics)
        )



import os
import unittest

from lyricsgenius.types import Song
from lyricsgenius.utils import clean_str
from tests import genius


class TestSong(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up Song tests...\n")

        cls.artist_name = "Andy Shauf"
        cls.song_title = "begin again"  # Lowercase is intentional
        cls.album = "The Party"
        cls.year = "2016-05-20"
        cls.song = genius.search_song(cls.song_title, cls.artist_name)
        genius.remove_section_headers = True
        cls.song_trimmed = genius.search_song(cls.song_title, cls.artist_name)

    def test_song(self):
        msg = "The returned object is not an instance of the Song class."
        self.assertIsInstance(self.song, Song, msg)

    def test_title(self):
        msg = "The returned song title does not match the title of the requested song."
        self.assertEqual(clean_str(self.song.title), clean_str(self.song_title), msg)

    def test_artist(self):
        # The returned artist name does not match the artist of the requested song.
        self.assertEqual(self.song.artist, self.artist_name)

    def test_lyrics_raw(self):
        lyrics = "[Verse 1]"
        self.assertTrue(self.song.lyrics.startswith(lyrics))

    def test_lyrics_no_section_headers(self):
        lyrics = "Begin again\nThis time you should take a bow at the"
        self.assertTrue(self.song_trimmed.lyrics.startswith(lyrics))

    def test_result_is_lyrics(self):
        self.assertTrue(genius._result_is_lyrics(self.song.to_dict()))

    def test_saving_json_file(self):
        print("\n")
        extension = "json"
        msg = "Could not save {} file.".format(extension)
        expected_filename = "lyrics_save_test_file." + extension
        filename = expected_filename.split(".")[0]

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving json file
        self.song.save_lyrics(filename=filename, extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting json file (now that file is written)
        try:
            self.song.save_lyrics(
                filename=expected_filename, extension=extension, overwrite=True
            )
            os.remove(expected_filename)
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
            os.remove(expected_filename)

    def test_saving_txt_file(self):
        print("\n")
        extension = "txt"
        msg = "Could not save {} file.".format(extension)
        expected_filename = "lyrics_save_test_file." + extension
        filename = expected_filename.split(".")[0]

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving txt file
        self.song.save_lyrics(filename=filename, extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting txt file (now that file is written)
        try:
            self.song.save_lyrics(
                filename=filename, extension=extension, overwrite=True
            )
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
        os.remove(expected_filename)



import os
import unittest

from lyricsgenius.types import Artist
from lyricsgenius.utils import sanitize_filename
from tests import genius


class TestArtist(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up Artist tests...\n")

        cls.artist_name = "The Beatles"
        cls.new_song = "Paperback Writer"
        cls.max_songs = 2
        cls.artist = genius.search_artist(cls.artist_name, max_songs=cls.max_songs)

    def test_artist(self):
        msg = "The returned object is not an instance of the Artist class."
        self.assertIsInstance(self.artist, Artist, msg)

    def test_correct_artist_name(self):
        msg = "Returned artist name does not match searched artist."
        name = "Queen"
        result = genius.search_artist(name, max_songs=1).name
        self.assertEqual(name, result, msg)

    def test_zero_songs(self):
        msg = "Songs were downloaded even though 0 songs was requested."
        name = "Queen"
        result = genius.search_artist(name, max_songs=0).songs
        self.assertEqual(0, len(result), msg)

    def test_name(self):
        msg = "The artist object name does not match the requested artist name."
        self.assertEqual(self.artist.name, self.artist_name, msg)

    def test_add_song_from_same_artist(self):
        msg = "The new song was not added to the artist object."
        self.artist.add_song(genius.search_song(self.new_song, self.artist_name))
        self.assertEqual(self.artist.num_songs, self.max_songs + 1, msg)

    def test_song(self):
        msg = "Song was not in artist's songs."
        song = self.artist.song(self.new_song)
        self.assertIsNotNone(song, msg)

    def test_add_song_from_different_artist(self):
        msg = "A song from a different artist was incorrectly allowed to be added."
        self.artist.add_song(genius.search_song("These Days", "Jackson Browne"))
        self.assertEqual(self.artist.num_songs, self.max_songs, msg)

    def test_artist_with_includes_features(self):
        # The artist did not get songs returned that they were featured in.
        name = "Swae Lee"
        result = genius.search_artist(name, max_songs=1, include_features=True)
        result = result.songs[0].artist
        self.assertNotEqual(result, name)

    def determine_filenames(self, extension):
        expected_filenames = []
        for song in self.artist.songs:
            fn = "lyrics_{name}_{song}.{ext}".format(
                name=self.artist.name, song=song.title, ext=extension
            )
            fn = sanitize_filename(fn.lower().replace(" ", ""))
            expected_filenames.append(fn)
        return expected_filenames

    def test_saving_json_file(self):
        print("\n")
        extension = "json"
        msg = "Could not save {} file.".format(extension)
        expected_filename = (
            "Lyrics_" + self.artist.name.replace(" ", "") + "." + extension
        )

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving json file
        self.artist.save_lyrics(extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting json file (now that file is written)
        try:
            self.artist.save_lyrics(extension=extension, overwrite=True)
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
        os.remove(expected_filename)

    def test_saving_txt_file(self):
        print("\n")
        extension = "txt"
        msg = "Could not save {} file.".format(extension)
        expected_filename = (
            "Lyrics_" + self.artist.name.replace(" ", "") + "." + extension
        )

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving txt file
        self.artist.save_lyrics(extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting txt file (now that file is written)
        try:
            self.artist.save_lyrics(extension=extension, overwrite=True)
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
        os.remove(expected_filename)



import unittest

from lyricsgenius import PublicAPI
from tests import genius

client = PublicAPI()


class TestAlbumMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up album methods tests...\n")

        cls.album_id = 104614

    def test_album(self):
        msg = "Album ID did not match."
        r = client.album(self.album_id)
        self.assertEqual(r["album"]["id"], self.album_id, msg)

    def test_albums_charts(self):
        msg = "Album charts were empty."
        r = client.albums_charts()
        self.assertTrue("chart_items" in r, msg)

    def test_album_comments(self):
        msg = "Album comments were empty."
        r = client.album_comments(self.album_id)
        self.assertTrue("comments" in r, msg)

    def test_album_cover_arts(self):
        msg = "Album cover arts were empty."
        r = client.album_cover_arts(self.album_id)
        self.assertTrue("cover_arts" in r, msg)

    def test_album_leaderboard(self):
        msg = "Album leaderboard was empty."
        r = client.album_leaderboard(self.album_id)
        self.assertTrue("leaderboard" in r, msg)

    def test_album_tracks(self):
        msg = "Album tracks were empty."
        r = client.album_tracks(self.album_id)
        self.assertTrue("tracks" in r, msg)


class TestAnnotationMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up annotation methods tests...\n")

        cls.annotation_id = 10225840

    def test_annotation(self):
        msg = "annotation ID did not match."
        r = client.annotation(self.annotation_id)
        self.assertEqual(r["annotation"]["id"], self.annotation_id, msg)

    def test_annotation_edits(self):
        msg = "annotation edits were empty."
        r = client.annotation_edits(self.annotation_id)
        self.assertTrue("versions" in r, msg)

    def test_annotation_comments(self):
        msg = "annotation comments were empty."
        r = client.annotation_comments(self.annotation_id)
        self.assertTrue("comments" in r, msg)


class TestArticleMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up article methods tests...\n")

        cls.article_id = 11880

    def test_article(self):
        msg = "article ID did not match."
        r = client.article(self.article_id)
        self.assertEqual(r["article"]["id"], self.article_id, msg)

    def test_article_comments(self):
        msg = "article comments were empty."
        r = client.article_comments(self.article_id)
        self.assertTrue("comments" in r, msg)

    def test_latest_articles(self):
        msg = "latest articles were empty."
        r = client.latest_articles()
        self.assertTrue("editorial_placements" in r, msg)


class TestArtistMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up artist methods tests...\n")

        cls.artist_id = 1665

    def test_artist(self):
        r = client.artist(self.artist_id)
        self.assertEqual(r["artist"]["id"], self.artist_id)

    def test_artist_activity(self):
        r = client.artist_activity(self.artist_id)
        self.assertTrue("line_items" in r)

    def test_artist_albums(self):
        r = client.artist_albums(self.artist_id)
        self.assertTrue("albums" in r)

    # def test_artist_contribution_opportunities(self):
    #     r = client.artist_contribution_opportunities(self.artist_id)
    #     self.assertIsNotNone(r.get('contribution_opportunities'))

    def test_artist_followers(self):
        r = client.artist_followers(self.artist_id)
        self.assertTrue("followers" in r)

    def test_artist_leaderboard(self):
        r = client.artist_leaderboard(self.artist_id)
        self.assertTrue("leaderboard" in r)

    def test_artist_songs(self):
        r = client.artist_songs(self.artist_id)
        self.assertTrue("songs" in r)

    def test_search_artist_songs(self):
        r = client.search_artist_songs(self.artist_id, "test")
        self.assertTrue("songs" in r)


class TestCoverArtMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up cover arts methods tests...\n")

        cls.album_id = 104614

    def test_cover_arts(self):
        r = client.cover_arts(self.album_id)
        self.assertTrue("cover_arts" in r)


class TestDiscussionMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up discussion methods tests...\n")

    #        cls.discussion_id = 123
    #
    #    def test_discussion(self):
    #        r = client.discussion(self.discussion_id)
    #        self.assertEqual(r['discussion']['id'], self.discussion_id)
    #
    #    def test_discussion_replies(self):
    #        r = client.discussion_replies(self.discussion_id)
    #        self.assertTrue("forum_posts" in r)

    def test_discussions(self):
        r = client.discussions()
        self.assertTrue("discussions" in r)


class TestLeaderboardMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up leaerboard methods tests...\n")

    def test_leaderboard(self):
        r = client.leaderboard()
        self.assertTrue("leaderboard" in r)

    def test_charts(self):
        r = client.charts()
        self.assertTrue("chart_items" in r)


class TestQuestionMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up question methods tests...\n")

        cls.album_id = 104614

    def test_questions(self):
        r = client.questions(self.album_id)
        self.assertIsNotNone(r.get("questions"))


class TestReferentMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up referent methods tests...\n")

        cls.web_page_id = 10347
        cls.referent_ids = [20793764, 20641014]

    def test_referent(self):
        r = client.referent(self.referent_ids)
        self.assertTrue(str(self.referent_ids[0]) in r["referents"])
        self.assertTrue(str(self.referent_ids[1]) in r["referents"])

    def test_referents(self):
        r = client.referents(web_page_id=self.web_page_id)
        self.assertIsNotNone(r.get("referents"))


class TestSearchMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up search methods tests...\n")

        cls.search_term = "test"

    def test_search(self):
        r = client.search(self.search_term)
        self.assertIsNotNone(r["hits"])

    def test_search_albums(self):
        r = client.search_albums(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "album")

    def test_search_articles(self):
        r = client.search_articles(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "article")

    def test_search_artists(self):
        r = client.search_artists(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "artist")

    def test_search_lyrics(self):
        r = client.search_lyrics(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "lyric")

    def test_search_songs(self):
        r = client.search_songs(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "song")

    def test_search_users(self):
        r = client.search_users(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "user")

    def test_search_videos(self):
        r = client.search_videos(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "video")

    def test_search_all(self):
        r = client.search_all(self.search_term)
        self.assertEqual(r["sections"][0]["type"], "top_hit")


class TestSongMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up song methods tests...\n")

        cls.song_id = 378195

    def test_song(self):
        r = client.song(self.song_id)
        self.assertEqual(r["song"]["id"], self.song_id)

    def test_song_activity(self):
        r = client.song_activity(self.song_id)
        self.assertTrue("line_items" in r)

    def test_song_comments(self):
        r = client.song_comments(self.song_id)
        self.assertTrue("comments" in r)

    def test_song_contributors(self):
        r = client.song_contributors(self.song_id)
        self.assertTrue("contributors" in r)


class TestUserMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up user methods tests...\n")

        cls.user_id = 1

    def test_user(self):
        r = client.user(self.user_id)
        self.assertEqual(r["user"]["id"], self.user_id)

    def test_user_accomplishments(self):
        r = client.user_accomplishments(self.user_id)
        self.assertTrue("accomplishments" in r)

    def test_user_following(self):
        r = client.user_following(self.user_id)
        self.assertTrue("followed_users" in r)

    def test_user_followers(self):
        r = client.user_followers(self.user_id)
        self.assertTrue("followers" in r)

    def test_user_contributions(self):
        r = client.user_contributions(self.user_id)
        self.assertTrue("contribution_groups" in r)

    def test_user_annotations(self):
        r = client.user_annotations(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "annotation")

    def test_user_articles(self):
        r = client.user_articles(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "article")

    def test_user_pyongs(self):
        r = client.user_pyongs(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "pyong")

    def test_user_questions_and_answers(self):
        r = client.user_questions_and_answers(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "answer")

    def test_user_suggestions(self):
        r = client.user_suggestions(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "comment")

    def test_user_transcriptions(self):
        r = client.user_transcriptions(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "song")

    def test_user_unreviewed(self):
        r = client.user_unreviewed(self.user_id)
        type = r["contribution_groups"][0]["contribution_type"]
        self.assertEqual(type, "annotation")


class TestVideoMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up video methods tests...\n")

        cls.video_id = 18681

    def test_video(self):
        r = client.video(self.video_id)
        self.assertEqual(r["video"]["id"], self.video_id)

    def test_videos(self):
        r = client.videos(video_id=self.video_id, series=True)
        self.assertTrue("video_lists" in r)

        r = client.videos(video_id=self.video_id)
        self.assertTrue("videos" in r)


class TestMiscMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up misc methods tests...\n")

        # cls.line_item_id = 146262999
        cls.annotation_id = 10225840

    # def test_line_item(self):
    #    r = client.line_item(self.line_item_id)
    #    self.assertTrue("line_item" in r)

    def test_page_data_album(self):
        album_path = "/albums/Eminem/Music-to-be-murdered-by"

        page_data = genius.page_data(album=album_path)
        self.assertTrue("page_data" in page_data)

    def test_page_data_song(self):
        artist = client.artist(1665)
        artist_slug = artist["artist"]["slug"]

        song = genius.song(4558484)
        song_path = song["song"]["path"]

        page_data = genius.page_data(artist=artist_slug, song=song_path)
        self.assertTrue("page_data" in page_data)

    def test_voters(self):
        r = client.voters(annotation_id=self.annotation_id)
        self.assertTrue("voters" in r)



import os
import platform
import time
from json.decoder import JSONDecodeError

import requests
from requests.exceptions import HTTPError, Timeout


class Sender(object):
    """Sends requests to Genius."""

    # Create a persistent requests connection
    API_ROOT = "https://api.genius.com/"
    PUBLIC_API_ROOT = "https://genius.com/api/"
    WEB_ROOT = "https://genius.com/"

    def __init__(
        self,
        access_token=None,
        response_format="plain",
        timeout=5,
        sleep_time=0.2,
        retries=0,
        public_api_constructor=False,
        user_agent="",
        proxy=None,
    ):
        self._session = requests.Session()
        user_agent_root = f"{platform.system()} {platform.release()}; Python {platform.python_version()}"
        self._session.headers = {
            "application": "LyricsGenius",
            "User-Agent": f"({user_agent}) ({user_agent_root})"
            if user_agent
            else user_agent_root,
        }
        if proxy:
            self._session.proxies = proxy
        if access_token is None:
            access_token = os.environ.get("GENIUS_ACCESS_TOKEN")

        if public_api_constructor:
            self.authorization_header = {}
        else:
            if not access_token or not isinstance(access_token, str):
                raise TypeError("Invalid token")
            self.access_token = "Bearer " + access_token
            self.authorization_header = {"authorization": self.access_token}

        self.response_format = response_format.lower()
        self.timeout = timeout
        self.sleep_time = sleep_time
        self.retries = retries

    def _make_request(
        self, path, method="GET", params_=None, public_api=False, web=False, **kwargs
    ):
        """Makes a request to Genius."""
        if public_api:
            uri = self.PUBLIC_API_ROOT
            header = None
        elif web:
            uri = self.WEB_ROOT
            header = None
        else:
            uri = self.API_ROOT
            header = self.authorization_header
        uri += path

        params_ = params_ if params_ else {}

        # Make the request
        response = None
        tries = 0
        while response is None and tries <= self.retries:
            tries += 1
            try:
                response = self._session.request(
                    method,
                    uri,
                    timeout=self.timeout,
                    params=params_,
                    headers=header,
                    **kwargs,
                )
                response.raise_for_status()
            except Timeout as e:
                error = "Request timed out:\n{e}".format(e=e)
                if tries > self.retries:
                    raise Timeout(error) from e
            except HTTPError as e:
                error = get_description(e)
                if response.status_code < 500 or tries > self.retries:
                    raise HTTPError(response.status_code, error) from e

            # Enforce rate limiting
            time.sleep(self.sleep_time)

        if web:
            return response.text
        elif response.status_code == 200:
            res = response.json()
            return res.get("response", res)
        elif response.status_code == 204:
            return 204
        else:
            raise AssertionError(
                "Response status code was neither 200, nor 204! It was {}".format(
                    response.status_code
                )
            )


def get_description(e):
    error = str(e)
    try:
        res = e.response.json()
    except JSONDecodeError:
        res = {}
    description = (
        res["meta"]["message"] if res.get("meta") else res.get("error_description")
    )
    error += "\n{}".format(description) if description else ""
    return error



import webbrowser
from urllib.parse import urlencode

from .api import Sender
from .errors import InvalidStateError
from .utils import parse_redirected_url


class OAuth2(Sender):
    """Genius OAuth2 authorization flow.

    Using this class you can authenticate a user,
    and get their token.

    Args:
        client_id (:obj:`str`): Client ID
        redirect_uri (:obj:`str`): Whitelisted redirect URI.
        client_secret (:obj:`str`, optional): Client secret.
        scope (:obj:`tuple` | :obj:`"all"`, optional): Token privileges.
        state (:obj:`str`, optional): Request state.
        client_only_app (:obj:`bool`, optional): `True` to use the client-only
            authorization flow, otherwise `False`.

    Raises:
        AssertionError: If neither :obj:`client_secret`, nor
            :obj:`client_only_app` is supplied.

    """

    auth_url = "https://api.genius.com/oauth/authorize"
    token_url = "https://api.genius.com/oauth/token"

    def __init__(
        self,
        client_id,
        redirect_uri,
        client_secret=None,
        scope=None,
        state=None,
        client_only_app=False,
    ):
        super().__init__()
        msg = (
            "You must provide a client_secret "
            "if you intend to use the full code exchange."
            "\nIf you meant to use the client-only flow, "
            "set the client_only_app parameter to True."
        )
        assert any([client_secret, client_only_app]), msg
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        if scope == "all":
            scope = ("me", "create_annotation", "manage_annotation", "vote")
        self.scope = scope if scope else ()
        self.state = state
        self.flow = "token" if client_only_app else "code"
        self.client_only_app = client_only_app

    @property
    def url(self):
        """Returns the URL you redirect the user to.

        You can use this property to get a URL that when opened on the user's
        device, shows Genius's authorization page where user clicks *Agree*
        to give your app access, and then Genius redirects user back to your
        redirect URI.

        """
        payload = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "response_type": self.flow,
        }
        if self.scope:
            payload["scope"] = " ".join(self.scope)
        if self.state:
            payload["state"] = self.state
        return OAuth2.auth_url + "?" + urlencode(payload)

    def get_user_token(self, code=None, url=None, state=None, **kwargs):
        """Gets a user token using the url or the code parameter..

        If you supply value for :obj:`code`, this method will use the value of the
        :obj:`code` parameter to request a token from Genius.

        If you use the :method`client_only_app` and supplt the redirected URL,
        it will already have the token.
        You could pass the URL to this method or parse it yourself.

        If you provide a :obj:`state` the method will also compare
        it to the initial state and will raise an exception if
        they're not equal.

        Args:
            code (:obj:`str`): 'code' parameter of redirected URL.
            url (:obj:`str`): Redirected URL (used in client-only apps)
            state (:obj:`str`): state parameter of redirected URL (only
                provide if you want to compare with initial :obj:`self.state`)
            **kwargs: keywords for the POST request.
        returns:
            :obj:`str`: User token.

        """
        assert any([code, url]), "You must pass either `code` or `url`."

        if state is not None and self.state != state:
            raise InvalidStateError("States do not match.")

        if code:
            payload = {
                "code": code,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "redirect_uri": self.redirect_uri,
                "grant_type": "authorization_code",
                "response_type": "code",
            }
            url = OAuth2.token_url.replace("https://api.genius.com/", "")
            res = self._make_request(url, "POST", data=payload, **kwargs)
            token = res["access_token"]
        else:
            token = parse_redirected_url(url, self.flow)
        return token

    def prompt_user(self):
        """Prompts current user for authentication.

        Opens a web browser for you to log in with Genius.
        Prompts to paste the URL after logging in to parse the
        *token* URL parameter.

        returns:
            :obj:`str`: User token.

        """

        url = self.url
        print("Opening browser for Genius login...")
        webbrowser.open(url)
        redirected = input("Please paste redirect URL: ").strip()

        if self.flow == "token":
            token = parse_redirected_url(redirected, self.flow)
        else:
            code = parse_redirected_url(redirected, self.flow)
            token = self.get_user_token(code)

        return token

    @classmethod
    def client_only_app(cls, client_id, redirect_uri, scope=None, state=None):
        """Returns an OAuth2 instance for a client-only app.

        Args:
            client_id (:obj:`str`): Client ID.
            redirect_uri (:obj:`str`): Whitelisted redirect URI.
            scope (:obj:`tuple` | :obj:`"all"`, optional): Token privilages.
            state (:obj:`str`, optional): Request state.

        returns:
            :class:`OAuth2`

        """
        return cls(
            client_id=client_id,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
            client_only_app=True,
        )

    @classmethod
    def full_code_exchange(
        cls, client_id, redirect_uri, client_secret, scope=None, state=None
    ):
        """Returns an OAuth2 instance for a full-code exchange app.

        Args:
            client_id (:obj:`str`): Client ID.
            redirect_uri (:obj:`str`): Whitelisted redirect URI.
            client_secret (:obj:`str`): Client secret.
            scope (:obj:`tuple` | :obj:`"all"`, optional): Token privilages.
            state (:obj:`str`, optional): Request state.

        returns:
            :class:`OAuth2`

        """
        return cls(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            scope=scope,
            state=state,
        )

    def __repr__(self):
        return (
            "{name}("
            "flow={flow!r}, "
            "scope={scope!r}, "
            "state={state!r}, "
            "client_only_app={client_only_app!r})"
        ).format(
            name=self.__class__.__name__,
            flow=self.flow,
            scope=self.scope,
            state=self.state,
            client_only_app=self.client_only_app,
        )



import unittest

from lyricsgenius.utils import (
    auth_from_environment,
    parse_redirected_url,
    sanitize_filename,
)
from tests import genius


class TestUtils(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up utils tests...\n")

    def test_sanitize_filename(self):
        raw = "B@ad File#_name"
        cleaned = "Bad File_name"
        r = sanitize_filename(raw)
        self.assertEqual(r, cleaned)

    def test_parse_redirected_url(self):
        redirected = "https://example.com/callback?code=test"
        flow = "code"
        code = "test"
        r = parse_redirected_url(redirected, flow)
        self.assertEqual(r, code)

        redirected = "https://example.com/callback#access_token=test"
        flow = "token"
        code = "test"
        r = parse_redirected_url(redirected, flow)
        self.assertEqual(r, code)

    def test_auth_from_environment(self):
        credentials = auth_from_environment()
        self.assertTrue(len(credentials) == 3)
        self.assertTrue(all(credentials))

    @classmethod
    def tearDownClass(cls):
        genius._session.close()



class AnnotationMethods(object):
    """Annotation methods of the public API."""

    def annotation(self, annotation_id, text_format=None):
        """Gets data for a specific annotation.

        Args:
            annotation_id (:obj:`int`): Genius annotation ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "annotations/{}".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def annotation_edits(self, annotation_id, text_format=None):
        """Gets the edits on annotation (its versions).

        Args:
            annotation_id (:obj:`int`): Genius annotation ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "annotations/{}/versions".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def annotation_comments(
        self, annotation_id, per_page=None, page=None, text_format=None
    ):
        """Gets the comments on an annotation.

        Args:
            annotation_id (:obj:`int`): Genius annotation ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "annotations/{}/comments".format(annotation_id)
        params = {
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)



# LyricsGenius
# copyright 2025 John W. R. Miller
# See LICENSE for details.

"""Artist object"""

from ..utils import safe_unicode
from .base import BaseEntity


class Artist(BaseEntity):
    """An artist with songs from the Genius.com database."""

    def __init__(self, client, json_dict):
        # Artist Constructor
        body = json_dict
        super().__init__(body["id"])
        self._body = body
        self._client = client
        self.songs = []
        self.num_songs = len(self.songs)

        self.api_path = body["api_path"]
        self.header_image_url = body["header_image_url"]
        self.image_url = body["image_url"]
        # self.iq = body['iq']
        self.is_meme_verified = body["is_meme_verified"]
        self.is_verified = body["is_verified"]
        self.name = body["name"]
        self.url = body["url"]

    def __len__(self):
        return len(self.songs)

    def add_song(self, new_song, verbose=True, include_features=False):
        """Adds a song to the Artist.

        This method adds a new song to the artist object. It checks
        if the song is already in artist's songs and whether the
        song's artist is the same as the `Artist` object.

        Args:
            new_song (:class:`Song <lyricsgenius.types.Song>`): Song to be added.
            verbose (:obj:`bool`, optional): prints operation result.
            include_features (:obj:`bool`, optional): If True, includes tracks
                featuring the artist.

        Returns:
            :obj:`int`: 0 for success and 1 for failure.

        Examples:
            .. code:: python

                genius = Genius(token)
                artist = genius.search_artist('Andy Shauf', max_songs=3)

                # Way 1
                song = genius.search_song('To You', artist.name)
                artist.add_song(song)

                # Way 2
                artist.add_song('To You')

        """
        if isinstance(new_song, str):
            new_song = self._client.search_song(new_song)
            if new_song is None:
                return None
        if any([song.title == new_song.title for song in self.songs]):
            if verbose:
                print(
                    "{s} already in {a}, not adding song.".format(
                        s=safe_unicode(new_song.title), a=safe_unicode(self.name)
                    )
                )
            return None
        if new_song.artist == self.name or (
            include_features and any(new_song._body.get("featured_artists", []))
        ):
            self.songs.append(new_song)
            self.num_songs += 1
            return new_song
        if verbose:
            print(
                "Can't add song by {b}, artist must be {a}.".format(
                    b=safe_unicode(new_song.artist), a=safe_unicode(self.name)
                )
            )
        return None

    def song(self, song_name):
        """Gets the artist's song.

        If the song is in the artist's songs, returns the song. Otherwise searches
        Genius for the song and then returns the song.

        Args:
            song_name (:obj:`str`): name of the song.
                the result is returned as a string.

        Returns:
            :obj:`Song <types.Song>` \\|‌ :obj:`None`: If it can't find the song,
            returns *None*.

        """
        for song in self.songs:
            if song.title == song_name:
                return song

        song = self._client.search_song(song_name, self.name)
        return song

    def to_dict(self):
        body = super().to_dict()
        body["songs"] = [song.to_dict() for song in self.songs]
        return body

    def to_json(self, filename=None, sanitize=True, ensure_ascii=True):
        data = self.to_dict()
        return super().to_json(
            data=data, filename=filename, sanitize=sanitize, ensure_ascii=ensure_ascii
        )

    def to_text(self, filename=None, sanitize=True):
        data = "\n\n".join(
            f"[Song {n}: {song.title}]\n{song.lyrics}"
            for n, song in enumerate(self.songs, start=1)
        ).strip()
        return super().to_text(data=data, filename=filename, sanitize=sanitize)

    def save_lyrics(
        self,
        filename=None,
        extension="json",
        overwrite=False,
        ensure_ascii=True,
        sanitize=True,
        verbose=True,
    ):
        # Determine the filename
        if filename is None:
            filename = "Lyrics_" + self.name.replace(" ", "")

        return super().save_lyrics(
            filename=filename,
            extension=extension,
            overwrite=overwrite,
            ensure_ascii=ensure_ascii,
            sanitize=sanitize,
            verbose=verbose,
        )

    def __str__(self):
        """Return a string representation of the Artist object."""
        msg = "{name}, {num} songs".format(name=self.name, num=self.num_songs)
        msg = msg[:-1] if self.num_songs == 1 else msg
        return msg



import unittest

from tests import genius


class TestAPI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up API tests...\n")

    def test_account(self):
        msg = (
            "No user detail was returned. "
            "Are you sure you're using a user access token?"
        )
        r = genius.account()
        self.assertTrue("user" in r, msg)

    def test_annotation(self):
        msg = "Returned annotation API path is different than expected."
        id_ = 10225840
        r = genius.annotation(id_)
        real = r["annotation"]["api_path"]
        expected = "/annotations/10225840"
        self.assertEqual(real, expected, msg)

    def test_manage_annotation(self):
        example_text = "The annotation"
        new_annotation = genius.create_annotation(
            example_text, "https://example.com", "illustrative examples", title="test"
        )["annotation"]
        msg = "Annotation text did not match the one that was passed."
        self.assertEqual(new_annotation["body"]["plain"], example_text, msg)

        try:
            example_text_two = "Updated annotation"
            r = genius.update_annotation(
                new_annotation["id"],
                example_text_two,
                "https://example.com",
                "illustrative examples",
                title="test",
            )["annotation"]
            msg = "Updated annotation text did not match the one that was passed."
            self.assertEqual(r["body"]["plain"], example_text_two, msg)

            r = genius.upvote_annotation(11828417)
            msg = "Upvote was not registered."
            self.assertTrue(r is not None, msg)

            r = genius.downvote_annotation(11828417)
            msg = "Downvote was not registered."
            self.assertTrue(r is not None, msg)

            r = genius.unvote_annotation(11828417)
            msg = "Vote was not removed."
            self.assertTrue(r is not None, msg)
        finally:
            msg = "Annotation was not deleted."
            r = genius.delete_annotation(new_annotation["id"])
            self.assertEqual(r, 204, msg)

    def test_referents_web_page(self):
        msg = "Returned referent API path is different than expected."
        id_ = 10347
        r = genius.referents(web_page_id=id_)
        real = r["referents"][0]["api_path"]
        expected = "/referents/11828416"
        self.assertTrue(real == expected, msg)

    def test_referents_no_inputs(self):
        # Must supply `song_id`, `web_page_id`, or `created_by_id`.
        with self.assertRaises(AssertionError):
            genius.referents()

    def test_referents_invalid_input(self):
        # Method should prevent inputs for both song and web_pag ID.
        with self.assertRaises(AssertionError):
            genius.referents(song_id=1, web_page_id=1)

    def test_web_page(self):
        msg = "Returned web page API path is different than expected."
        url = "https://docs.genius.com"
        r = genius.web_page(raw_annotatable_url=url)
        real = r["web_page"]["api_path"]
        expected = "/web_pages/10347"
        self.assertEqual(real, expected, msg)



class VideoMethods(object):
    """Video methods of the public API."""

    def video(self, video_id, text_format=None):
        """Gets data for a specific video.

        Args:
            video_id (:obj:`int`): Genius video ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "videos/{}".format(video_id)
        params = {"text_format": text_format or self.response_format}

        return self._make_request(path=endpoint, params_=params, public_api=True)

    def videos(
        self,
        album_id=None,
        article_id=None,
        song_id=None,
        video_id=None,
        per_page=None,
        page=None,
        series=False,
    ):
        """Gets the videos of an album, article or song or the featured videos.

        Args:
            album_id (:obj:`int`, optional): Genius album ID
            article_id (:obj:`int`, optional): Genius article ID
            song_id (:obj:`int`, optional): Genius song ID
            video_id (:obj:`int`, optional): Genius video ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            series (:obj:`bool`, optional): If set to `True`, returns episodes
            of Genius original video series that the item has been mentioned in.

        Returns:
            :obj:`dict`

        Note:
            If you specify no album, article or song, the method will return
            a series of videos. In this case, if `series=True`, the results
            will be the videos in the *VIDEOS* section on the homepage. But
            if `series=False`, the method returns another set of videos that
            we are not sure what they are at the moment.

        """
        msg = (
            "Pass only one of `album_id`, `article_id`, `song_id` and `video_id`."
            ", not more than one."
        )
        condition = (
            sum([bool(album_id), bool(article_id), bool(song_id), bool(video_id)]) == 1
        )
        assert condition, msg

        if series:
            endpoint = "video_lists"
        else:
            endpoint = "videos"

        params = {"per_page": per_page, "page": page}

        if album_id:
            params["album_id"] = album_id
        elif article_id:
            params["article_id"] = article_id
        elif song_id:
            params["song_id"] = song_id
        elif video_id:
            params["video_id"] = video_id

        return self._make_request(path=endpoint, params_=params, public_api=True)



import os
import unittest
from unittest.mock import MagicMock, patch

from lyricsgenius import OAuth2
from lyricsgenius.errors import InvalidStateError

client_id = os.environ["GENIUS_CLIENT_ID"]
client_secret = os.environ["GENIUS_CLIENT_SECRET"]
redirect_uri = os.environ["GENIUS_REDIRECT_URI"]


def mocked_requests_post(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

        def raise_for_status(self):
            if self.status_code > 300:
                raise ConnectionError

    method, url = args[0], args[1]
    data = kwargs["data"]
    code = data.get("code")
    data_client_id = data.get("client_id")
    data_client_secret = data.get("client_secret")
    data_redirect_uri = data.get("redirect_uri")
    data_grant_type = data.get("grant_type")
    data_response_type = data.get("response_type")

    if (
        method == "POST"
        and url == OAuth2.token_url
        and code == "some_code"
        and data_client_id == client_id
        and data_client_secret == client_secret
        and data_redirect_uri == redirect_uri
        and data_grant_type == "authorization_code"
        and data_response_type == "code"
    ):
        return MockResponse({"access_token": "test"}, 200)

    return MockResponse(None, 403)


class TestOAuth2(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up OAuth2 tests...\n")

    def test_init(self):
        with self.assertRaises(AssertionError):
            OAuth2(client_id, redirect_uri)

        scope = ("me", "create_annotation", "manage_annotation", "vote")
        auth = OAuth2(client_id, redirect_uri, client_secret, scope="all")
        self.assertEqual(auth.scope, scope)

    @patch("requests.Session.request", side_effect=mocked_requests_post)
    def test_get_user_token_code_flow(self, mock_post):
        # full code exchange flow

        state = "some_state"
        code = "some_code"
        code_flow_token = "test"

        auth = OAuth2.full_code_exchange(
            client_id, redirect_uri, client_secret, scope="all", state=state
        )

        r = auth.get_user_token(code=code, state=state)
        self.assertEqual(r, code_flow_token)

    def test_get_user_token_token_flow(self):
        state = "some_state"
        token_flow_token = "test"
        redirected_url = "{}#access_token=test".format(redirect_uri)

        auth = OAuth2.client_only_app(client_id, redirect_uri, scope="all", state=state)

        r = auth.get_user_token(url=redirected_url)
        self.assertEqual(r, token_flow_token)

    def test_get_user_token_invalid_state(self):
        state = "state_1"
        auth = OAuth2.full_code_exchange(
            client_id, redirect_uri, client_secret, scope="all", state=state
        )

        returned_code = "some_code"
        returned_state = "state_2"
        with self.assertRaises(InvalidStateError):
            auth.get_user_token(code=returned_code, state=returned_state)

    def test_get_user_token_no_parameter(self):
        state = "some_state"
        auth = OAuth2.full_code_exchange(
            client_id, redirect_uri, client_secret, scope="all", state=state
        )

        with self.assertRaises(AssertionError):
            auth.get_user_token()

    def test_prompt_user(self):
        auth = OAuth2(client_id, redirect_uri, client_secret, scope="all")
        token = "test"
        current_module = "lyricsgenius.auth"

        input_ = MagicMock(return_value="http://example.com?code=some_code")
        with (
            patch(current_module + ".webbrowser", MagicMock()),
            patch(current_module + ".input", input_),
            patch(current_module + ".print", MagicMock()),
            patch("requests.Session.request", side_effect=mocked_requests_post),
        ):
            r = auth.prompt_user()

        self.assertEqual(r, token)



import json
import os
from abc import ABC, abstractmethod

from ..utils import safe_unicode, sanitize_filename


class BaseEntity(ABC):
    """Base class for types."""

    def __init__(self, id):
        self.id = id

    @abstractmethod
    def save_lyrics(
        self,
        filename,
        extension="json",
        overwrite=False,
        ensure_ascii=True,
        sanitize=True,
        verbose=True,
    ):
        """Save Song(s) lyrics and metadata to a JSON or TXT file.

        If the extension is 'json' (the default), the lyrics will be saved
        alongside the song's information. Take a look at the example below.

        Args:
            filename (:obj:`str`, optional): Output filename, a string.
                If not specified, the result is returned as a string.
            extension (:obj:`str`, optional): Format of the file (`json` or `txt`).
            overwrite (:obj:`bool`, optional): Overwrites preexisting file if `True`.
                Otherwise prompts user for input.
            ensure_ascii (:obj:`bool`, optional): If ensure_ascii is true
                (the default), the output is guaranteed to have all incoming
                non-ASCII characters escaped.
            sanitize (:obj:`bool`, optional): Sanitizes the filename if `True`.
            verbose (:obj:`bool`, optional): prints operation result.

        Warning:
            If you set :obj:`sanitize` to `False`, the file name may contain
            invalid characters, and thefore cause the saving to fail.

        """
        extension = extension.lstrip(".").lower()
        msg = "extension must be JSON or TXT"
        assert (extension == "json") or (extension == "txt"), msg

        # Determine the filename
        for ext in [".txt", ".TXT", ".json", ".JSON"]:
            if ext in filename:
                filename = filename.replace(ext, "")
                break
        filename += "." + extension
        filename = sanitize_filename(filename) if sanitize else filename

        # Check if file already exists
        write_file = False
        if overwrite or not os.path.isfile(filename):
            write_file = True
        elif verbose:
            msg = "{} already exists. Overwrite?\n(y/n): ".format(filename)
            if input(msg).lower() == "y":
                write_file = True

        # Exit if we won't be saving a file
        if not write_file:
            if verbose:
                print("Skipping file save.\n")
            return

        # Save the lyrics to a file
        if extension == "json":
            self.to_json(filename, ensure_ascii=ensure_ascii, sanitize=sanitize)
        else:
            self.to_text(filename, sanitize=sanitize)

        if verbose:
            print("Wrote {}.".format(safe_unicode(filename)))

        return None

    @abstractmethod
    def to_dict(self):
        """Converts the object to a dictionary."""
        return self._body.copy()

    @abstractmethod
    def to_json(self, data, filename=None, sanitize=True, ensure_ascii=True):
        """Converts the object to a json string.

        Args:
            filename (:obj:`str`, optional): Output filename, a string.
                If not specified, the result is returned as a string.
            sanitize (:obj:`bool`, optional): Sanitizes the filename if `True`.
            ensure_ascii (:obj:`bool`, optional): If ensure_ascii is true
              (the default), the output is guaranteed to have all incoming
              non-ASCII characters escaped.

        Returns:
            :obj:`str` \\|‌ :obj:`None`: If :obj:`filename` is `None`,
            returns the lyrics as a plain string, otherwise `None`.

        Warning:
            If you set :obj:`sanitize` to `False`, the file name may contain
            invalid characters, and therefore cause the saving to fail.

        """
        data = self.to_dict()

        # Return the json string if no output path was specified
        if not filename:
            return json.dumps(data, indent=1, ensure_ascii=ensure_ascii)

        # Save Song object to a json file
        filename = sanitize_filename(filename) if sanitize else filename
        with open(filename, "w", encoding="utf-8") as ff:
            json.dump(data, ff, indent=4, ensure_ascii=ensure_ascii)
        return None

    @abstractmethod
    def to_text(self, data, filename=None, sanitize=True):
        """Converts song(s) lyrics to a single string.

        Args:
            filename (:obj:`str`, optional): Output filename, a string.
                If not specified, the result is returned as a string.
            sanitize (:obj:`bool`, optional): Sanitizes the filename if `True`.

        Returns:
            :obj:`str` \\|‌ :obj:`None`: If :obj:`filename` is `None`,
            returns the lyrics as a plain string. Otherwise `None`.

        Warning:
            If you set :obj:`sanitize` to `False`, the file name may contain
            invalid characters, and therefore cause the saving to fail.

        """
        # Return the lyrics as a string if no `filename` was specified
        if not filename:
            return data

        # Save song lyrics to a text file
        filename = sanitize_filename(filename) if sanitize else filename
        with open(filename, "w", encoding="utf-8") as ff:
            ff.write(data)
        return None

    def __repr__(self):
        name = self.__class__.__name__
        attrs = [x for x in list(self.__dict__.keys()) if not x.startswith("_")]
        attrs = ", ".join(attrs[:2])
        return "{}({}, ...)".format(name, attrs)


class Stats(object):
    """Stats of an item.

    Note:
        The values passed to this class are inconsistent,
        and therefore need to be set dynamically.
        Use the built-in ``dir()`` function to
        see the available attributes.
        You could also access the stats by the dictionary
        annotation. For example:

        .. code:: python

            values = song.to_dict()
            print(values['stats'])

    """

    def __init__(self, json_dict):
        for key, value in json_dict.items():
            setattr(self, key, value)

    def __repr__(self):
        name = self.__class__.__name__
        attrs = ", ".join(list(self.__dict__.keys()))
        return "{}({!r})".format(name, attrs)


# class EntityWithLyrics(ABC, BaseEntity):
#    """Entity that has lyrics."""
#
#    def __init__(self, **kwargs):
#        super().__init__(**kwargs)
#
#    @abstractmethod
#    def save_lyrics(self):
#        pass



import unittest

from requests.exceptions import HTTPError

from tests import genius


class TestAPIBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up API base tests...\n")

    def test_http_error_handler(self):
        status_code = None
        try:
            genius.annotation(0)
        except HTTPError as e:
            status_code = e.args[0]

        self.assertEqual(status_code, 404)



import os
import unittest
import warnings

from lyricsgenius.types import Album
from tests import genius


class TestAlbum(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up Album tests...\n")
        warnings.simplefilter("ignore", ResourceWarning)
        cls.album_name = "The Party"
        cls.artist_name = "Andy Shauf"
        cls.num_tracks = 10
        cls.album = genius.search_album(cls.album_name, cls.artist_name)

    def test_type(self):
        self.assertIsInstance(self.album, Album)

    def test_album_name(self):
        self.assertEqual(self.album.name, self.album_name)

    def test_album_artist(self):
        self.assertEqual(self.album.artist.name, self.artist_name)

    def test_tracks(self):
        self.assertEqual(len(self.album.tracks), self.num_tracks)

    def test_saving_json_file(self):
        print("\n")
        extension = "json"
        msg = "Could not save {} file.".format(extension)
        expected_filename = (
            "Lyrics_" + self.album.name.replace(" ", "") + "." + extension
        )

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving json file
        self.album.save_lyrics(extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting json file (now that file is written)
        try:
            self.album.save_lyrics(extension=extension, overwrite=True)
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
        os.remove(expected_filename)

    def test_saving_txt_file(self):
        print("\n")
        extension = "txt"
        msg = "Could not save {} file.".format(extension)
        expected_filename = (
            "Lyrics_" + self.album.name.replace(" ", "") + "." + extension
        )

        # Remove the test file if it already exists
        if os.path.isfile(expected_filename):
            os.remove(expected_filename)

        # Test saving txt file
        self.album.save_lyrics(extension=extension, overwrite=True)
        self.assertTrue(os.path.isfile(expected_filename), msg)

        # Test overwriting txt file (now that file is written)
        try:
            self.album.save_lyrics(extension=extension, overwrite=True)
        except Exception:
            self.fail("Failed {} overwrite test".format(extension))
        os.remove(expected_filename)



class AlbumMethods(object):
    """Album methods of the public API."""

    def album(self, album_id, text_format=None):
        """Gets data for a specific album.

        Args:
            album_id (:obj:`int`): Genius album ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Examples:
            .. code:: python

                genius = Genius(token)
                song = genius.search_song(378195)
                album_id = song['album']['id']
                album = genius.album(album_id)
                print(album['name'])

        """
        endpoint = "albums/{}".format(album_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def albums_charts(
        self,
        time_period="day",
        chart_genre="all",
        per_page=None,
        page=None,
        text_format=None,
    ):
        """Gets the album charts.

        Alias for :meth:`charts() <PublicAPI.charts>`.

        Args:
            time_period (:obj:`str`, optional): Time period of the results
                ('day', 'week', 'month' or 'all_time').
            chart_genre (:obj:`str`, optional): The genre of the results.
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        return self.charts(
            time_period=time_period,
            chart_genre=chart_genre,
            per_page=per_page,
            page=page,
            text_format=text_format,
            type_="albums",
        )

    def album_comments(self, album_id, per_page=None, page=None, text_format=None):
        """Gets the comments on an album page.

        Args:
            album_id (:obj:`int`): Genius album ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "albums/{}/comments".format(album_id)
        params = {
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def album_cover_arts(self, album_id, text_format=None):
        """Gets cover arts of a specific album.

        Alias for :meth:`cover_arts <PublicAPI.cover_arts>`.

        Args:
            album_id (:obj:`int`): Genius album ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain'). Defines the text
                formatting for the annotation of the cover arts,
                if there are any.

        Returns:
            :obj:`dict`

        Examples:
            Downloading album's cover art:

            .. code:: python

                import requests

                genius = Genius(token)
                res = genius.album_cover_arts(104614)
                cover_art = requests.get(res['cover_arts'][0]['image_url'])

        """
        return self.cover_arts(album_id=album_id, text_format=text_format)

    def album_leaderboard(self, album_id, per_page=None, page=None):
        """Gets the leaderboard of an album.

        This method returns the album's top contributors.

        Args:
            album_id (:obj:`int`): Genius album ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "albums/{}/leaderboard".format(album_id)
        params = {"per_page": per_page, "page": page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def album_tracks(self, album_id, per_page=None, page=None, text_format=None):
        """Gets the tracks of a specific album.

        Args:
            album_id (:obj:`int`): Genius album ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "albums/{}/tracks".format(album_id)
        params = {
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)



class MiscMethods(object):
    """Miscellaneous Methods"""

    def line_item(self, line_item_id, text_format=None):
        """Gets data for a specific line item.

        Args:
            line_item_id (:obj:`int`): Genius line item ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Warning:
            This method requires a logged in user and will raise
            ``NotImplementedError``.

        """
        raise NotImplementedError("This action requires a logged in user.")

        endpoint = "line_items/{}".format(line_item_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def page_data(self, album=None, song=None, artist=None):
        """Gets page data of an item.

        If you want the page data of a song, you must supply
        song and artist. But if you want the page data of an album,
        you only have to supply the album.

        Page data will return all possible values for the album/song and
        the lyrics in HTML format if the item is a song!
        Album page data will contian album info and tracks info as well.

        Args:
            album (:obj:`str`, optional): Album path
                (e.g. '/albums/Eminem/Music-to-be-murdered-by')
            song (:obj:`str`, optional): Song path
                (e.g. '/Sia-chandelier-lyrics')
            artist (:obj:`str`, optional): Artist slug. (e.g. 'Andy-shauf')

        Returns:
            :obj:`dict`

        Warning:
            Some albums/songs either don't have page data or
            their page data path can't be infered easily from
            the artist slug and their API path. So make sure to
            use this method with a try/except clause that catches
            404 errors. Check out the example below.


        Examples:
            Getting the lyrics of a song from its page data

            .. code:: python

                from lyricsgenius import Genius, PublicAPI
                from bs4 import BeautifulSoup
                from requests import HTTPError

                genius = Genius(token)
                public = PublicAPI()

                # We need the PublicAPI to get artist's slug
                artist = public.artist(1665)
                artist_slug = artist['artist']['slug']

                # The rest can be done using Genius
                song = genius.song(4558484)
                song_path = song['song']['path']

                try:
                    page_data = genius.page_data(artist=artist_slug, song=song_path)
                except HTTPError as e:
                    print("Couldn't find page data {}".format(e.status_code))
                    page_data = None

                if page_data is not None:
                    lyrics_html = page_data['page_data']['lyrics_data']['body']['html']
                    lyrics_text = BeautifulSoup(lyrics_html, 'html.parser').get_text()

        """
        assert any([album, song]), "You must pass either song or album."
        if song:
            assert all([song, artist]), "You must pass artist."

        if album:
            endpoint = "page_data/album"
            page_type = "albums"
            item_path = album.replace("/albums/", "")
        else:
            endpoint = "page_data/song"
            page_type = "songs"

            # item path becomes something like: Artist/Song
            item_path = (
                song[1:].replace(artist + "-", artist + "/").replace("-lyrics", "")
            )

        page_path = "/{page_type}/{item_path}".format(
            page_type=page_type, item_path=item_path
        )
        params = {"page_path": page_path}

        return self._make_request(endpoint, params_=params, public_api=True)

    def voters(
        self, annotation_id=None, answer_id=None, article_id=None, comment_id=None
    ):
        """Gets the voters of an item.

        You must supply one of :obj:`annotation_id`, :obj:`answer_id`, :obj:`article_id`
        or :obj:`comment_id`.

        Args:
            annotation_id (:obj:`int`, optional): Genius annotation ID
            answer_id (:obj:`int`, optional): Genius answer ID
            article_id (:obj:`int`, optional): Genius article ID
            comment_id (:obj:`int`, optional): Genius comment ID

        Returns:
            :obj:`dict`

        """
        msg = "Must supply `annotation_id`, `answer_id`, `comment_id` or `article_id`"
        assert any([annotation_id, answer_id, article_id, comment_id]), msg
        msg = (
            "Pass only one of "
            "`annotation_id`, `answer_id`, `article_id` or `comment_id`"
            ", not more than one."
        )
        condition = (
            sum(
                [
                    bool(annotation_id),
                    bool(answer_id),
                    bool(article_id),
                    bool(comment_id),
                ]
            )
            == 1
        )
        assert condition, msg

        endpoint = "voters"

        params = {}
        if annotation_id:
            params["annotation_id"] = annotation_id
        elif answer_id:
            params["answer_id"] = answer_id
        elif article_id:
            params["article_id"] = article_id
        elif comment_id:
            params["comment_id"] = comment_id
        return self._make_request(path=endpoint, params_=params, public_api=True)



# GeniusAPI
# John W. R. Miller
# See LICENSE for details
"""A library that provides a Python interface to the Genius API"""

import sys

assert sys.version_info[0] == 3, "LyricsGenius requires Python 3."
from lyricsgenius.api import API, PublicAPI
from lyricsgenius.auth import OAuth2
from lyricsgenius.genius import Genius
from lyricsgenius.utils import auth_from_environment



import unittest

from tests import genius


class TestEndpoints(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up Endpoint tests...\n")

        cls.search_term = "Ezra Furman"
        cls.song_title_only = "99 Problems"
        cls.tag = genius.tag("pop")

    def test_search_song(self):
        artist = "Jay-Z"
        # Empty response
        response = genius.search_song("")
        self.assertIsNone(response)

        # Pass no title and ID
        with self.assertRaises(AssertionError):
            genius.search_song()

        # Search by song ID
        response = genius.search_song(song_id=1)
        self.assertIsNotNone(response)

        # Exact match exact search
        response = genius.search_song(self.song_title_only)
        self.assertTrue(response.title.lower() == self.song_title_only.lower())

        # Song with artist name
        response = genius.search_song(self.song_title_only, artist)
        self.assertTrue(response.title.lower() == self.song_title_only.lower())

        # Spaced out search
        response = genius.search_song("  \t 99  \t \t\tProblems   ", artist)
        self.assertTrue(response.title.lower() == self.song_title_only.lower())

        # No title match because of artist
        response = genius.search_song(self.song_title_only, artist="Drake")
        self.assertFalse(response.title.lower() == self.song_title_only.lower())

    def test_song_annotations(self):
        msg = "Incorrect song annotation response."
        r = sorted(genius.song_annotations(1))
        real = r[0][0]
        expected = "(I’m at bat)"
        self.assertEqual(real, expected, msg)

    def test_tag_results(self):
        r = self.tag

        self.assertEqual(r["next_page"], 2)
        self.assertEqual(len(r["hits"]), 20)

    def test_tag_first_result(self):
        artists = ["Luis Fonsi", "Daddy Yankee"]
        featured_artists = ["Justin Bieber"]
        song_title = "Despacito (Remix)"
        title_with_artists = (
            "Despacito (Remix) by Luis Fonsi & Daddy Yankee (Ft. Justin Bieber)"
        )
        url = "https://genius.com/Luis-fonsi-and-daddy-yankee-despacito-remix-lyrics"

        first_song = self.tag["hits"][0]

        self.assertEqual(artists, first_song["artists"])
        self.assertEqual(featured_artists, first_song["featured_artists"])
        self.assertEqual(song_title, first_song["title"])
        self.assertEqual(title_with_artists, first_song["title_with_artists"])
        self.assertEqual(url, first_song["url"])


class TestLyrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("\n---------------------\nSetting up lyrics tests...\n")

        cls.song_url = "https://genius.com/Andy-shauf-begin-again-lyrics"
        cls.song_id = 2885745
        cls.lyrics_ending = (
            "[Outro]"
            "\nNow I’m kicking leaves"
            "\nCursing the one that I love and the one I don't"
            "\nI wonder who you're thinking of"
        )

    def test_lyrics_with_url(self):
        lyrics = genius.lyrics(song_url=self.song_url)
        self.assertTrue(lyrics.endswith(self.lyrics_ending))

    def test_lyrics_with_id(self):
        lyrics = genius.lyrics(self.song_id)
        self.assertTrue(lyrics.endswith(self.lyrics_ending))



from .base import Sender
from .public_methods import (
    AlbumMethods,
    AnnotationMethods,
    ArticleMethods,
    ArtistMethods,
    CoverArtMethods,
    DiscussionMethods,
    LeaderboardMethods,
    MiscMethods,
    QuestionMethods,
    ReferentMethods,
    SearchMethods,
    SongMethods,
    UserMethods,
    VideoMethods,
)


class API(Sender):
    """Genius API.

    The :obj:`API` class is in charge of making all the requests
    to the developers' API (api.genius.com)
    Use the methods of this class if you already have information
    such as song ID to make direct requests to the API. Otherwise
    the :class:`Genius` class provides a friendlier front-end
    to search and retrieve data from Genius.com.

    All methods of this class are available through the :class:`Genius` class.

    Args:
        access_token (:obj:`str`): API key provided by Genius.
        response_format (:obj:`str`, optional): API response format (dom, plain, html).
        timeout (:obj:`int`, optional): time before quitting on response (seconds).
        sleep_time (:obj:`str`, optional): time to wait between requests.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.
        user_agent (:obj:`str`, optional): User agent for the request header.
        proxy (:obj:`dict[str, str]`, optional): Proxy settings.

    Attributes:
        response_format (:obj:`str`, optional): API response format (dom, plain, html).
        timeout (:obj:`int`, optional): time before quitting on response (seconds).
        sleep_time (:obj:`str`, optional): time to wait between requests.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.

    Returns:
        :class:`API`: An object of the `API` class.

    """

    def __init__(
        self,
        access_token,
        response_format="plain",
        timeout=5,
        sleep_time=0.2,
        retries=0,
        user_agent="",
        proxy=None,
    ):
        super().__init__(
            access_token=access_token,
            response_format=response_format,
            timeout=timeout,
            sleep_time=sleep_time,
            retries=retries,
            user_agent=user_agent,
            proxy=proxy,
        )

    def account(self, text_format=None):
        """Gets details about the current user.

        Requires scope: :obj:`me`.

        Args:
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "account"
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, params_=params)

    def annotation(self, annotation_id, text_format=None):
        """Gets data for a specific annotation.

        Args:
            annotation_id (:obj:`int`): annotation ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        params = {"text_format": text_format or self.response_format}
        endpoint = "annotations/{id}".format(id=annotation_id)
        return self._make_request(endpoint, params_=params)

    def create_annotation(
        self,
        text,
        raw_annotatable_url,
        fragment,
        before_html=None,
        after_html=None,
        canonical_url=None,
        og_url=None,
        title=None,
        text_format=None,
    ):
        """Creates an annotation for a web page.

        Requires scope: :obj:`create_annotation`.

        Args:
            text (:obj:`str`): Annotation text in Markdown format.
            raw_annotatable_url (:obj:`str`): The original URL of the page.
            fragment (:obj:`str`): The highlighted fragment (the referent).
            before_html (:obj:`str`, optional): The HTML before the highlighted fragment
                (prefer up to 200 characters).
            after_html (:obj:`str`, optional): The HTML after the highlighted fragment
                (prefer up to 200 characters).
            canonical_url (:obj:`str`, optional): The href property of the
                :obj:`<link rel="canonical">` tag on the page.
            og_url (:obj:`str`, optional): The content property of the
                :obj:`<meta property="og:url">` tag on the page.
            title (:obj:`str`, optional): The title of the page.
            text_format (:obj:`str`, optional): Text format of the response
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`: The annotation.

        Examples:
            .. code:: python

                genius = Genius(token)
                new = genius.update_annotation('The annotation',
                'https://example.com', 'illustrative examples', title='test')
                print(new['id'])

        """
        msg = "Must supply `canonical_url`, `og_url`, or `title`."
        assert any([canonical_url, og_url, title]), msg

        endpoint = "annotations"
        params = {"text_format": text_format or self.response_format}
        payload = {
            "annotation": {"body": {"markdown": text}},
            "referent": {
                "raw_annotatable_url": raw_annotatable_url,
                "fragment": fragment,
                "context_for_display": {
                    "before_html": before_html if before_html else None,
                    "after_html": after_html if after_html else None,
                },
            },
            "web_page": {
                "canonical_url": canonical_url if canonical_url else None,
                "og_url": og_url if og_url else None,
                "title": title if title else None,
            },
        }
        return self._make_request(
            path=endpoint, method="POST", params_=params, json=payload
        )

    def delete_annotation(self, annotation_id):
        """Deletes an annotation created by the authenticated user.

        Requires scope: :obj:`manage_annotation`.

        Args:
            annotation_id (:obj:`int`): Annotation ID.

        Returns:
            :obj:`int`: 204 - which is the response's status code

        """
        endpoint = "annotations/{}".format(annotation_id)
        return self._make_request(path=endpoint, method="DELETE")

    def downvote_annotation(self, annotation_id, text_format=None):
        """Downvotes an annotation.

        Requires scope: :obj:`vote`.

        Args:
            annotation_id (:obj:`int`): Annotation ID.
            text_format (:obj:`str`, optional): Text format of the response
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`: The annotation.

        """
        endpoint = "annotations/{}/downvote".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, method="PUT", params_=params)

    def unvote_annotation(self, annotation_id, text_format=None):
        """Removes user's vote for the annotation.

        Requires scope: :obj:`vote`.

        Args:
            annotation_id (:obj:`int`): Annotation ID.
            text_format (:obj:`str`, optional): Text format of the response
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`: The annotation.

        """
        endpoint = "annotations/{}/unvote".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, method="PUT", params_=params)

    def update_annotation(
        self,
        annotation_id,
        text,
        raw_annotatable_url,
        fragment,
        before_html=None,
        after_html=None,
        canonical_url=None,
        og_url=None,
        title=None,
        text_format=None,
    ):
        """Updates an annotation created by the authenticated user.

        Requires scope: :obj:`manage_annotation`.

        Args:
            annotation_id (:obj:`int`): ID of the annotation that will be updated.
            text (:obj:`str`): Annotation text in Markdown format.
            raw_annotatable_url (:obj:`str`): The original URL of the page.
            fragment (:obj:`str`): The highlighted fragment (the referent).
            before_html (:obj:`str`, optional): The HTML before the highlighted fragment
                (prefer up to 200 characters).
            after_html (:obj:`str`, optional): The HTML after the highlighted fragment
                (prefer up to 200 characters).
            canonical_url (:obj:`str`, optional): The href property of the
                :obj:`<link rel="canonical">` tag on the page.
            og_url (:obj:`str`, optional): The content property of the
                :obj:`<meta property="og:url">` tag on the page.
            title (:obj:`str`, optional): The title of the page.
            text_format (:obj:`str`, optional): Text format of the response
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`: The annotation.

        """
        msg = "Must supply `canonical_url`, `og_url`, or `title`."
        assert any([canonical_url, og_url, title]), msg

        endpoint = "annotations/{}".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        payload = {
            "annotation": {"body": {"markdown": text}},
            "referent": {
                "raw_annotatable_url": raw_annotatable_url,
                "fragment": fragment,
                "context_for_display": {
                    "before_html": before_html if before_html else None,
                    "after_html": after_html if after_html else None,
                },
            },
            "web_page": {
                "canonical_url": canonical_url if canonical_url else None,
                "og_url": og_url if og_url else None,
                "title": title if title else None,
            },
        }
        return self._make_request(
            path=endpoint, method="PUT", params_=params, json=payload
        )

    def upvote_annotation(self, annotation_id, text_format=None):
        """Upvotes an annotation.

        Requires scope: :obj:`vote`.

        Args:
            annotation_id (:obj:`int`): Annotation ID.
            text_format (:obj:`str`, optional): Text format of the response
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`: The annotation.

        """
        endpoint = "annotations/{}/upvote".format(annotation_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(path=endpoint, method="PUT", params_=params)

    def artist(self, artist_id, text_format=None):
        """Gets data for a specific artist.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Examples:
            .. code:: python

                genius = Genius(token)
                artist = genius.artist(380491)
                print(artist['name'])

        """
        params = {"text_format": text_format or self.response_format}
        endpoint = "artists/{id}".format(id=artist_id)
        return self._make_request(endpoint, params_=params)

    def artist_songs(self, artist_id, per_page=None, page=None, sort="title"):
        """Gets artist's songs.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            sort (:obj:`str`, optional): Sorting preference.
                Either based on 'title', 'popularity' or 'release_date'.
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        Examples:
            .. code:: python

                # getting all artist songs based on popularity
                genius = Genius(token)
                page = 1
                songs = []
                while page:
                    request = genius.artist_songs(380491,
                                                  sort='popularity',
                                                  per_page=50,
                                                  page=page)
                songs.extend(request['songs'])
                page = request['next_page']
                least_popular_song = songs[-1]['title']


                # getting songs 11-15
                songs = genius.artist_songs(380491, per_page=5, page=3)

        """
        endpoint = "artists/{id}/songs".format(id=artist_id)

        params = {"sort": sort, "per_page": per_page, "page": page}
        return self._make_request(endpoint, params_=params)

    def referents(
        self,
        song_id=None,
        web_page_id=None,
        created_by_id=None,
        per_page=None,
        page=None,
        text_format=None,
    ):
        """Gets item's referents

        Args:
            song_id (:obj:`int`, optional): song ID
            web_page_id (:obj:`int`, optional): web page ID
            created_by_id (:obj:`int`, optional): User ID of the contributer
                who created the annotation(s).
            per_page (:obj:`int`, optional): Number of results to
                return per page. It can't be more than 50.
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Note:
            You may pass only one of :obj:`song_id` and
            :obj:`web_page_id`, not both.

        Examples:
            .. code:: python

                # getting all verified annotations of a song (artist annotations)
                genius = Genius(token)
                request = genius.referents(song_id=235729,
                                           per_page=50)
                verified = [y for x in request['referents']
                            for y in x['annotations'] if y['verified']]

        """
        msg = "Must supply `song_id`, `web_page_id`, or `created_by_id`."
        assert any([song_id, web_page_id, created_by_id]), msg
        msg = "Pass only one of `song_id` and `web_page_id`, not both."
        assert bool(song_id) ^ bool(web_page_id), msg

        # Construct the URI
        endpoint = "referents"
        params = {"text_format": text_format or self.response_format}
        params = {
            "song_id": song_id,
            "web_page_id": web_page_id,
            "created_by_id": created_by_id,
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(endpoint, params_=params)

    def search_songs(self, search_term, per_page=None, page=None):
        """Searches songs hosted on Genius.

        Args:
            search_term (:obj:`str`): A term to search on Genius.
            per_page (:obj:`int`, optional): Number of results to
                return per page. It can't be more than 5 for this method.
            page (:obj:`int`, optional): Number of the page.

        Returns:
            :obj:`dict`

        """
        endpoint = "search"
        params = {"q": search_term, "per_page": per_page, "page": page}
        return self._make_request(endpoint, params_=params)

    def song(self, song_id, text_format=None):
        """Gets data for a specific song.

        Args:
            song_id (:obj:`int`): Genius song ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Examples:
            .. code:: python

                genius = Genius(token)
                song = genius.song(2857381)
                print(song['full_title'])

        """
        endpoint = "songs/{id}".format(id=song_id)
        params = {"text_format": text_format or self.response_format}
        return self._make_request(endpoint, params_=params)

    def web_page(self, raw_annotatable_url=None, canonical_url=None, og_url=None):
        """Gets data for a specific web page.

        Args:
            raw_annotatable_url (:obj:`str`): The URL as it would appear in a browser.
            canonical_url (:obj:`str`): The URL as specified by an appropriate <link>
                tag in a page's <head>.
            og_url (:obj:`str`): The URL as specified by an og:url <meta> tag in
                a page's <head>.

        Returns:
            :obj:`dict`

        Examples:
            .. code:: python

                genius = Genius(token)
                webpage = genius.web_page('docs.genius.com')
                print(webpage['full_title'])

        Note:
            * Data is only available for pages that already have at
              least one annotation.
            * You must at least pass one argument to the method.
            * You can pass more than one or all arguments (provided they're the address
              of the same webpage).

        """
        msg = "Must supply `raw_annotatable_url`, `canonical_url`, or `og_url`."
        assert any([raw_annotatable_url, canonical_url, og_url]), msg

        endpoint = "web_pages/lookup"
        params = {
            "raw_annotatable_url": raw_annotatable_url,
            "canonical_url": canonical_url,
            "og_url": og_url,
        }
        return self._make_request(path=endpoint, params_=params)


class PublicAPI(
    Sender,
    AlbumMethods,
    AnnotationMethods,
    ArticleMethods,
    ArtistMethods,
    CoverArtMethods,
    DiscussionMethods,
    LeaderboardMethods,
    QuestionMethods,
    ReferentMethods,
    SearchMethods,
    SongMethods,
    UserMethods,
    VideoMethods,
    MiscMethods,
):
    """Genius public API.

    The :obj:`PublicAPI` class is in charge of making all the requests
    to the public API (genius.com/api)
    You can use this method without an access token since calls are made
    to the public API.

    All methods of this class are available through the :class:`Genius` class.

    Args:
        response_format (:obj:`str`, optional): API response format (dom, plain, html).
        timeout (:obj:`int`, optional): time before quitting on response (seconds).
        sleep_time (:obj:`str`, optional): time to wait between requests.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.

    Attributes:
        response_format (:obj:`str`, optional): API response format (dom, plain, html).
        timeout (:obj:`int`, optional): time before quitting on response (seconds).
        sleep_time (:obj:`str`, optional): time to wait between requests.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.

    Returns:
        :class:`PublicAPI`: An object of the `PublicAPI` class.

    """

    def __init__(
        self,
        response_format="plain",
        timeout=5,
        sleep_time=0.2,
        retries=0,
        user_agent="",
        **kwargs,
    ):
        # If PublicAPI was instantiated directly
        # there is no need for a token anymore
        public_api_constructor = False if self.__class__.__name__ == "Genius" else True

        # Genius PublicAPI Constructor
        super().__init__(
            response_format=response_format,
            timeout=timeout,
            sleep_time=sleep_time,
            retries=retries,
            public_api_constructor=public_api_constructor,
            user_agent=user_agent,
            **kwargs,
        )



# GeniusAPI
# John W. R. Miller
# See LICENSE for details

"""API documentation: https://docs.genius.com/"""

import json
import os
import re
import shutil
import time

from bs4 import BeautifulSoup

from .api import API, PublicAPI
from .types import Album, Artist, Song, Track
from .utils import clean_str, safe_unicode


class Genius(API, PublicAPI):
    """User-level interface with the Genius.com API and public API.

    Args:
        access_token (:obj:`str`, optional): API key provided by Genius.
        response_format (:obj:`str`, optional): API response format (dom, plain, html).
        timeout (:obj:`int`, optional): time before quitting on response (seconds).
        sleep_time (:obj:`str`, optional): time to wait between requests.
        verbose (:obj:`bool`, optional): Turn printed messages on or off.
        remove_section_headers (:obj:`bool`, optional): If `True`, removes [Chorus],
            [Bridge], etc. headers from lyrics.
        skip_non_songs (:obj:`bool`, optional): If `True`, attempts to
            skip non-songs (e.g. track listings).
        excluded_terms (:obj:`list`, optional): extra terms for flagging results
            as non-lyrics.
        replace_default_terms (:obj:`list`, optional): if True, replaces default
            excluded terms with user's. Default excluded terms are listed below.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.
        user_agent (:obj:`str`, optional): User agent for the request header.
        proxy (:obj:`dict[str, str]`, optional): Proxy settings.

    Attributes:
        verbose (:obj:`bool`, optional): Turn printed messages on or off.
        remove_section_headers (:obj:`bool`, optional): If `True`, removes [Chorus],
            [Bridge], etc. headers from lyrics.
        skip_non_songs (:obj:`bool`, optional): If `True`, attempts to
            skip non-songs (e.g. track listings).
        excluded_terms (:obj:`list`, optional): extra terms for flagging results
            as non-lyrics.
        replace_default_terms (:obj:`list`, optional): if True, replaces default
            excluded terms with user's.
        retries (:obj:`int`, optional): Number of retries in case of timeouts and
            errors with a >= 500 response code. By default, requests are only made once.

    Returns:
        :class:`Genius`

    Note:
        Default excluded terms are the following regular expressions:
        :obj:`track\\s?list`, :obj:`album art(work)?`, :obj:`liner notes`,
        :obj:`booklet`, :obj:`credits`, :obj:`interview`, :obj:`skit`,
        :obj:`instrumental`, and :obj:`setlist`.

    """

    default_terms = [
        "track\\s?list",
        "album art(work)?",
        "liner notes",
        "booklet",
        "credits",
        "interview",
        "skit",
        "instrumental",
        "setlist",
    ]

    def __init__(
        self,
        access_token=None,
        response_format="plain",
        timeout=5,
        sleep_time=0.2,
        verbose=True,
        remove_section_headers=False,
        skip_non_songs=True,
        excluded_terms=None,
        replace_default_terms=False,
        retries=0,
        user_agent="",
        proxy=None,
    ):
        # Genius Client Constructor
        super().__init__(
            access_token=access_token,
            response_format=response_format,
            timeout=timeout,
            sleep_time=sleep_time,
            retries=retries,
            user_agent=user_agent,
            proxy=proxy,
        )

        self.verbose = verbose
        self.remove_section_headers = remove_section_headers
        self.skip_non_songs = skip_non_songs

        excluded_terms = excluded_terms if excluded_terms is not None else []
        if replace_default_terms:
            self.excluded_terms = excluded_terms
        else:
            self.excluded_terms = self.default_terms.copy()
            self.excluded_terms.extend(excluded_terms)

    def lyrics(self, song_id=None, song_url=None, remove_section_headers=False):
        """Uses BeautifulSoup to scrape song info off of a Genius song URL

        You must supply either `song_id` or song_url`.

        Args:
            song_id (:obj:`int`, optional): Song ID.
            song_url (:obj:`str`, optional): Song URL.
            remove_section_headers (:obj:`bool`, optional):
                If `True`, removes [Chorus], [Bridge], etc. headers from lyrics.

        Returns:
            :obj:`str` \\|‌ :obj:`None`:
                :obj:`str` If it can find the lyrics, otherwise `None`

        Note:
            If you pass a song ID, the method will have to make an extra request
            to obtain the song's URL and scrape the lyrics off of it. So it's best
            to pass the method the song's URL if it's available.

            If you want to get a song's lyrics by searching for it,
            use :meth:`Genius.search_song` instead.

        Note:
            This method removes the song headers based on the value of the
            :attr:`Genius.remove_section_headers` attribute.

        """
        msg = "You must supply either `song_id` or `song_url`."
        assert any([song_id, song_url]), msg
        if song_url:
            path = song_url.replace("https://genius.com/", "")
        else:
            path = self.song(song_id)["song"]["path"][1:]

        # Scrape the song lyrics from the HTML
        html = BeautifulSoup(
            self._make_request(path, web=True).replace("<br/>", "\n"), "html.parser"
        )

        # Determine the class of the div
        divs = html.find_all(
            "div", class_=re.compile(r"^Lyrics-\w{2}.\w+.[1]|Lyrics__Container")
        )
        if divs is None or len(divs) <= 0:
            if self.verbose:
                print(
                    "Couldn't find the lyrics section. "
                    "Please report this if the song has lyrics.\n"
                    "Song URL: https://genius.com/{}".format(path)
                )
            return None

        lyrics = "\n".join([div.get_text() for div in divs])

        # Remove [Verse], [Bridge], etc.
        if self.remove_section_headers or remove_section_headers:
            lyrics = re.sub(r"(\[.*?\])*", "", lyrics)
            lyrics = re.sub("\n{2}", "\n", lyrics)  # Gaps between verses
        return lyrics.strip("\n")

    def _result_is_lyrics(self, song):
        """Returns False if result from Genius is not actually song lyrics.

        Sets the :attr:`lyricsgenius.Genius.excluded_terms` and
        :attr:`lyricsgenius.Genius.replace_default_terms` as instance variables
        within the Genius class.

        Args:
            song_title (:obj:`str`, optional): Title of the song.

        Returns:
            :obj:`bool`: `True` if none of the terms are found in the song title.

        Note:
            Default excluded terms are the following: 'track\\s?list',
            'album art(work)?', 'liner notes', 'booklet', 'credits',
            'interview', 'skit', 'instrumental', and 'setlist'.

        """
        if song["lyrics_state"] != "complete" or song.get("instrumental"):
            return False

        expression = r"".join(["({})|".format(term) for term in self.excluded_terms])
        expression = expression.strip("|")
        regex = re.compile(expression, re.IGNORECASE)
        return not regex.search(clean_str(song["title"]))

    def _get_item_from_search_response(self, response, search_term, type_, result_type):
        """Gets the desired item from the search results.

        This method tries to match the `hits` of the :obj:`response` to
        the :obj:`response_term`, and if it finds no match, returns the first
        appropriate hit if there are any.

        Args:
            response (:obj:`dict`): A response from
                :meth:‍‍‍‍`Genius.search_all` to go through.
            search_term (:obj:`str`): The search term to match with the hit.
            type_ (:obj:`str`): Type of the hit we're looking for (e.g. song, artist).
            result_type (:obj:`str`): The part of the hit we want to match
                (e.g. song title, artist's name).

        Returns:
            :obj:‍‍`str` \\|‌ :obj:`None`:
            - `None` if there is no hit in the :obj:`response`.
            - The matched result if matching succeeds.
            - The first hit if the matching fails.

        """

        # Convert list to dictionary
        top_hits = response["sections"][0]["hits"]

        # Check rest of results if top hit wasn't the search type
        sections = sorted(response["sections"], key=lambda sect: sect["type"] == type_)

        hits = [hit for hit in top_hits if hit["index"] == type_]
        hits.extend(
            [
                hit
                for section in sections
                for hit in section["hits"]
                if hit["index"] == type_
            ]
        )

        for hit in hits:
            item = hit["result"]
            if clean_str(item[result_type]) == clean_str(search_term):
                return item

        # If the desired type is song lyrics and none of the results matched,
        # return the first result that has lyrics
        if type_ == "song" and self.skip_non_songs:
            for hit in hits:
                song = hit["result"]
                if self._result_is_lyrics(song):
                    return song

        return hits[0]["result"] if hits else None

    def _result_is_match(self, result, title, artist=None):
        """Returns `True` if search result matches searched song."""
        result_title = clean_str(result["title"])
        title_is_match = result_title == clean_str(title)
        if not artist:
            return title_is_match
        result_artist = clean_str(result["primary_artist"]["name"])
        return title_is_match and result_artist == clean_str(artist)

    def song_annotations(self, song_id, text_format=None):
        """Return song's annotations with associated fragment in list of tuple.

        Args:
            song_id (:obj:`int`): song ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`list`: list of tuples(fragment, [annotations])

        Note:
            This method uses :meth:`Genius.referents`, but provides convenient
            access to fragments (annotated text) and the corresponding
            annotations (Some fragments may have more than one annotation,
            because sometimes both artists and Genius users annotate them).

        """
        referents = self.referents(song_id=song_id, text_format=text_format)

        all_annotations = []  # list of tuples(fragment, annotations[])
        for r in referents["referents"]:
            fragment = r["fragment"]
            annotations = []
            for a in r["annotations"]:
                annotations.append([x for x in a["body"].values()])
            all_annotations.append((fragment, annotations))
        return all_annotations

    def search_album(
        self, name=None, artist="", album_id=None, get_full_info=True, text_format=None
    ):
        """Searches for a specific album and gets its songs.

        You must pass either a :obj:`name` or an :obj:`album_id`.

        Args:
            name (:obj:`str`, optional): Album name to search for.
            artist (:obj:`str`, optional): Name of the artist.
            album_id (:obj:`int`, optional): Album ID.
            get_full_info (:obj:`bool`, optional): Get full info
                for the album (slower if no album_id present).
            text_format (:obj:`bool`, optional): If ``True``,
                gets each song's lyrics.

        Returns:
            :class:`Album <types.Album>` \\| :obj:`None`: On success,
            the album object is returned, otherwise `None`.

        Tip:
            Set :attr:`Genius.verbose` to `True` to read why the search fails.

        Examples:
            .. code:: python

                genius = Genius(token)
                album = genius.search_album("Andy Shauf", "The Party")
                print(album.name)

        """
        msg = "You must pass either a `name` or an `album_id`."
        assert any([name, album_id]), msg

        if self.verbose and name:
            if artist:
                print('Searching for "{s}" by {a}...'.format(s=name, a=artist))
            else:
                print('Searching for "{s}"...'.format(s=name))

        if album_id:
            album_info = self.album(album_id, text_format)["album"]
        else:
            search_term = "{s} {a}".format(s=name, a=artist).strip()
            response = self.search_all(search_term)
            album_info = self._get_item_from_search_response(
                response, name, type_="album", result_type="name"
            )

        # Exit search if there were no results returned from API
        # Otherwise, move forward with processing the search results
        if album_info is None:
            if self.verbose and name:
                print("No results found for: '{s}'".format(s=search_term))
            return None

        # If the album was searched, query the API using the album id so the full info can be retrieved
        if album_id is None and get_full_info:
            album_info.update(self.album(album_info["id"], text_format)["album"])

        # Set the album id to the value retrieved from the API
        album_id = album_info["id"]

        tracks = []
        next_page = 1

        # It's unlikely for an album to have >=50 songs,
        # but it's best to check
        while next_page:
            tracks_list = self.album_tracks(
                album_id=album_id, per_page=50, page=next_page, text_format=text_format
            )
            for track in tracks_list["tracks"]:
                song_info = track["song"]
                if song_info["lyrics_state"] == "complete" and not song_info.get(
                    "instrumental"
                ):
                    song_lyrics = self.lyrics(song_url=song_info["url"])
                else:
                    song_lyrics = ""

                track = Track(self, track, song_lyrics)
                tracks.append(track)

            next_page = tracks_list["next_page"]

        return Album(self, album_info, tracks)

    def search_song(self, title=None, artist="", song_id=None, get_full_info=True):
        """Searches for a specific song and gets its lyrics.

        You must pass either a :obj:`title` or a :obj:`song_id`.

        Args:
            title (:obj:`str`): Song title to search for.
            artist (:obj:`str`, optional): Name of the artist.
            get_full_info (:obj:`bool`, optional): Get full info for each song (slower).
            song_id (:obj:`int`, optional): Song ID.

        Returns:
            :class:`Song <types.Song>` \\| :obj:`None`: On success,
            the song object is returned, otherwise `None`.

        Tip:
            Set :attr:`Genius.verbose` to `True` to read why the search fails.

        Examples:
            .. code:: python

                genius = Genius(token)
                artist = genius.search_artist('Andy Shauf', max_songs=0)
                song = genius.search_song('Toy You', artist.name)
                # same as: song = genius.search_song('To You', 'Andy Shauf')
                print(song.lyrics)

        """
        msg = "You must pass either a `title` or a `song_id`."
        if title is None and song_id is None:
            assert any([title, song_id]), msg

        if self.verbose and title:
            if artist:
                print('Searching for "{s}" by {a}...'.format(s=title, a=artist))
            else:
                print('Searching for "{s}"...'.format(s=title))

        if song_id:
            song_info = self.song(song_id)["song"]
        else:
            search_term = "{s} {a}".format(s=title, a=artist).strip()
            search_response = self.search_all(search_term)
            song_info = self._get_item_from_search_response(
                search_response, title, type_="song", result_type="title"
            )

        # Exit search if there were no results returned from API
        # Otherwise, move forward with processing the search results
        if song_info is None:
            if self.verbose and title:
                print("No results found for: '{s}'".format(s=search_term))
            return None

        # Reject non-songs (Liner notes, track lists, etc.)
        # or songs with uncomplete lyrics (e.g. unreleased songs, instrumentals)
        if self.skip_non_songs and not self._result_is_lyrics(song_info):
            valid = False
        else:
            valid = True

        if not valid:
            if self.verbose:
                print("Specified song does not contain lyrics. Rejecting.")
            return None

        # Download full song info (an API call) unless told not to by user
        if song_id is None and get_full_info:
            song_info.update(self.song(song_info["id"])["song"])

        if song_info["lyrics_state"] == "complete" and not song_info.get(
            "instrumental"
        ):
            lyrics = self.lyrics(song_url=song_info["url"])
        else:
            lyrics = ""

        # Skip results when URL is a 404 or lyrics are missing
        if self.skip_non_songs and not lyrics:
            if self.verbose:
                print("Specified song does not have a valid lyrics. Rejecting.")
            return None

        # Return a Song object with lyrics if we've made it this far
        song = Song(self, song_info, lyrics)
        if self.verbose:
            print("Done.")
        return song

    def search_artist(
        self,
        artist_name,
        max_songs=None,
        sort="popularity",
        per_page=20,
        get_full_info=True,
        allow_name_change=True,
        artist_id=None,
        include_features=False,
    ):
        """Searches for a specific artist and gets their songs.

        This method looks for the artist by the name or by the
        ID if it's provided in ``artist_id``. It returrns an
        :class:`Artist <types.Artist>` object if the search is successful.
        If :obj:`allow_name_change` is True, the name of the artist is changed to the
        artist name on Genius.

        Args:
            artist_name (:obj:`str`): Name of the artist to search for.
            max_songs (obj:`int`, optional): Maximum number of songs to search for.
            sort (:obj:`str`, optional): Sort by 'title' or 'popularity'.
            per_page (:obj:`int`, optional): Number of results to return
                per search page. It can't be more than 50.
            get_full_info (:obj:`bool`, optional): Get full info for each song (slower).
            allow_name_change (:obj:`bool`, optional): If True, search attempts to
                switch to intended artist name.
            artist_id (:obj:`int`, optional): Allows user to pass an artist ID.
            include_features (:obj:`bool`, optional): If True, includes tracks
                featuring the artist.

        Returns:
            :class:`Artist <types.Artist>`: Artist object containing
            artist's songs.

        Examples:
            .. code:: python

                # printing the lyrics of all of the artist's songs
                genius = Genius(token)
                artist = genius.search_artist('Andy Shauf')
                for song in artist.songs:
                    print(song.lyrics)

            Visit :class:`Aritst <types.Artist>` for more examples.
        """

        def find_artist_id(search_term):
            """Finds the ID of the artist, returns the first
            result if none match the search term or returns
            ‍None‍‍ if there were not results

            """
            if self.verbose:
                print("Searching for songs by {0}...\n".format(search_term))

            # Perform a Genius API search for the artist
            found_artist = None
            response = self.search_all(search_term)
            found_artist = self._get_item_from_search_response(
                response, search_term, type_="artist", result_type="name"
            )

            # Exit the search if we couldn't find an artist by the given name
            if not found_artist:
                if self.verbose:
                    print("No results found for '{a}'.".format(a=search_term))
                return None
            # Assume the top search result is the intended artist
            return found_artist["id"]

        # Get the artist ID (or use the one supplied)
        artist_id = artist_id if artist_id else find_artist_id(artist_name)
        if not artist_id:
            return None

        artist_info = self.artist(artist_id)["artist"]
        found_name = artist_info["name"]
        if found_name != artist_name and allow_name_change:
            if self.verbose:
                print(
                    "Changing artist name to '{a}'".format(a=safe_unicode(found_name))
                )
            artist_name = found_name

        # Create the Artist object
        artist = Artist(self, artist_info)
        # Download each song by artist, stored as Song objects in Artist object
        page = 1
        reached_max_songs = True if max_songs == 0 else False
        while not reached_max_songs:
            songs_on_page = self.artist_songs(
                artist_id=artist_id,
                per_page=per_page,
                page=page,
                sort=sort,
            )

            # Loop through each song on page of search results
            for song_info in songs_on_page["songs"]:
                # Check if song is valid (e.g. contains lyrics)
                if self.skip_non_songs and not self._result_is_lyrics(song_info):
                    valid = False
                else:
                    valid = True

                # Reject non-song results (e.g. Linear Notes, Tracklists, etc.)
                if not valid:
                    if self.verbose:
                        s = song_info["title"]
                        print('"{s}" is not valid. Skipping.'.format(s=safe_unicode(s)))
                    continue

                # Create the Song object from lyrics and metadata
                if song_info["lyrics_state"] == "complete":
                    lyrics = self.lyrics(song_url=song_info["url"])
                else:
                    lyrics = ""
                if get_full_info:
                    new_info = self.song(song_info["id"])["song"]
                    song_info.update(new_info)
                song = Song(self, song_info, lyrics)

                # Attempt to add the Song to the Artist
                result = artist.add_song(
                    song, verbose=False, include_features=include_features
                )
                if result is not None and self.verbose:
                    print(
                        'Song {n}: "{t}"'.format(
                            n=artist.num_songs, t=safe_unicode(song.title)
                        )
                    )

                # Exit search if the max number of songs has been met
                reached_max_songs = max_songs and artist.num_songs >= max_songs
                if reached_max_songs:
                    if self.verbose:
                        print(
                            (
                                "\nReached user-specified song limit ({m}).".format(
                                    m=max_songs
                                )
                            )
                        )
                    break

            # Move on to next page of search results
            page = songs_on_page["next_page"]
            if page is None:
                break  # Exit search when last page is reached

        if self.verbose:
            print("Done. Found {n} songs.".format(n=artist.num_songs))
        return artist

    def save_artists(
        self, artists, filename="artist_lyrics", overwrite=False, ensure_ascii=True
    ):
        """Saves lyrics from multiple Artist objects as JSON object.

        Args:
            artists (:obj:`list`): List of :class:`Artist <types.Artist>`
                objects to save lyrics from.
            filename (:obj:`str`, optional): Name of the output file.
            overwrite (:obj:`bool`, optional): Overwrites preexisting file if `True`.
                Otherwise prompts user for input.
            ensure_ascii (:obj:`bool`, optional): If ensure_ascii is true
              (the default), the output is guaranteed to have all incoming
              non-ASCII characters escaped.

        Examples:
            .. code:: python

                genius = Genius(token)
                a = search_artist('Andy Shauf')
                b = search_artist('Queen', max_song=10)
                c = search_artist('The Beatles', max_songs=1)

                genius.save_artists(artists=[a, b, c], filename='abc', overwrite=True)

        """
        if isinstance(artists, Artist):
            artists = [artists]

        # Create a temporary directory for lyrics
        start = time.time()
        tmp_dir = "tmp_lyrics"
        if not os.path.isdir(tmp_dir):
            os.mkdir(tmp_dir)
            count = 0
        else:
            count = len(os.listdir(tmp_dir))

        # Check if file already exists
        if os.path.isfile(filename + ".json") and not overwrite:
            msg = "{f} already exists. Overwrite?\n(y/n): ".format(f=filename)
            if input(msg).lower() != "y":
                print("Leaving file in place. Exiting.")
                os.rmdir(tmp_dir)
                return

        # Extract each artist's lyrics in json format
        all_lyrics = {"artists": []}
        for n, artist in enumerate(artists):
            if isinstance(artist, Artist):
                all_lyrics["artists"].append({})
                f = "tmp_{n}_{a}".format(n=count + n, a=artist.name.replace(" ", ""))
                tmp_file = os.path.join(tmp_dir, f)
                if self.verbose:
                    print(tmp_file)
                all_lyrics["artists"][-1] = artist.save_lyrics(overwrite=True)

        # Save all of the lyrics
        with open(filename + ".json", "w", encoding="utf-8") as outfile:
            json.dump(all_lyrics, outfile, ensure_ascii=ensure_ascii)

        # Delete the temporary directory
        shutil.rmtree(tmp_dir)
        elapsed = (time.time() - start) / 60 / 60
        print("Time elapsed: {t} hours".format(t=elapsed))

    def tag(self, name, page=None):
        """Gets a tag's songs.

        This method parses HTML to extract the hits on the page, so it's
        slower than a normal API call which returns the results as JSON.

        Args:
            name (:obj:`int`): Name of the tag e.g. ``pop`` or ``r-b``.
                The name should be as it appears in the URL. For example,
                the name supplied for R&B is ``r-b`` as its URL is
                https://genius.com/tags/r-b.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`: A dictionary with the following keys:
            ``hits``: A list of dictionaries.
            ``next_page``: Either ``int`` or ``None``.

        Examples:
            .. code:: python

                # getting the lyrics of all the songs in the pop tag.
                genius = Genius(token)
                page = 1
                lyrics = []
                while page:
                    res = genius.tag('pop', page=page)
                    for hit in res['hits']:
                        song_lyrics = genius.lyrics(song_url=hit['url'])
                        lyrics.append(song_lyrics)
                    page = res['next_page']

        """
        path = "tags/{}/all".format(name)
        params = {"page": page}
        soup = BeautifulSoup(
            self._make_request(path, params_=params, web=True), "html.parser"
        )
        hits = []

        ul = soup.find("ul", class_="song_list")
        for li in ul.find_all("li"):
            url = li.a.attrs["href"]
            # Genius uses \xa0 in the HTML to add spaces
            song = [x.replace("\xa0", " ") for x in li.a.span.stripped_strings]
            title = song[0]
            artists = song[2].split(" & ")
            featured_artists = [name for name in song[4:-1] if len(name) > 1]
            title_with_artists = (
                li.a.find("span", class_="title_with_artists")
                .get_text()
                .strip()
                .replace("\xa0", " ")
            )

            hit = {
                "url": url,
                "title_with_artists": title_with_artists,
                "title": title,
                "artists": artists,
                "featured_artists": featured_artists,
            }
            hits.append(hit)

        res = {"hits": hits}
        page = page if page is not None else 1
        # Full pages contain 20 items
        res["next_page"] = page + 1 if len(hits) == 20 else None

        return res



class ArtistMethods(object):
    """Artist methods of the public API."""

    def artist(self, artist_id, text_format=None):
        """Gets data for a specific artist.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}".format(artist_id)
        params = {"text_format": text_format or self.response_format}

        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_activity(self, artist_id, per_page=None, page=None, text_format=None):
        """Gets activities on artist's songs.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/activity_stream/line_items".format(artist_id)
        params = {
            "per_page": per_page,
            "page": page,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_albums(self, artist_id, per_page=None, page=None):
        """Gets artist's albums.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/albums".format(artist_id)
        params = {"per_page": per_page, "page": page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_contribution_opportunities(
        self, artist_id, per_page=None, next_curosr=None, text_format=None
    ):
        """Gets contribution opportunities related to the artist.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            next_cursor (:obj:`int`, optional): Paginated offset
                (address of the next cursor).
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        Warning:
            This method requires a logged in user and will raise
            ``NotImplementedError``.

        """
        raise NotImplementedError("This action requires a logged in user.")
        endpoint = "artists/{}/contribution_opportunities".format(artist_id)
        params = {
            "per_page": per_page,
            "next_curosr": next_curosr,
            "text_format": text_format or self.response_format,
        }
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_followers(self, artist_id, per_page=None, page=None):
        """Gets artist's followers.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/followers".format(artist_id)
        params = {"per_page": per_page, "page": page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_leaderboard(self, artist_id, per_page=None, page=None):
        """Gets artist's top scholars.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/leaderboard".format(artist_id)
        params = {"per_page": per_page, "page": page}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def artist_songs(self, artist_id, per_page=None, page=None, sort="popularity"):
        """Gets artist's songs.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/songs".format(artist_id)
        params = {"per_page": per_page, "page": page, "sort": sort}
        return self._make_request(path=endpoint, params_=params, public_api=True)

    def search_artist_songs(
        self, artist_id, search_term, per_page=None, page=None, sort="popularity"
    ):
        """Searches artist's songs.

        Args:
            artist_id (:obj:`int`): Genius artist ID
            search_term (:obj:`str`): A term to search on Genius.
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            sort (:obj:`str`, optional): Sorting preference.
                ('title' or 'popularity')

        Returns:
            :obj:`dict`

        """
        endpoint = "artists/{}/songs/search".format(artist_id)
        params = {"q": search_term, "per_page": per_page, "page": page, "sort": sort}
        return self._make_request(path=endpoint, params_=params, public_api=True)



import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------

project = "lyricsgenius"
copyright = "2025, John W. R. Miller, Allerter"
author = "John W. R. Miller, Allerter"

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
]

exclude_patterns = ["build"]
master_doc = "index"
autosummary_generate = True
# -- Options for HTML output -------------------------------------------------

highlight_language = "python3"
html_theme = "sphinx_rtd_theme"

# -- Other -------------------------------------------------------------------

extlinks = {
    "issue": ("https://github.com/johnwmillr/LyricsGenius/issues/%s", "%s"),
    "commit": ("https://github.com/johnwmillr/LyricsGenius/commit/%s", "%s"),
}



#  Command line usage:
#    python3 -m lyricsgenius --help
#    python3 -m lyricsgenius song "Begin Again" "Andy Shauf" --save
#    python3 -m lyricsgenius artist "The Beatles" --max-songs 5 --save

import argparse
import os

from . import Genius
from .utils import safe_unicode


def main(args=None):
    msg = "Download song lyrics from Genius.com"
    parser = argparse.ArgumentParser(description=msg)
    parser.add_argument(
        "search_type",
        type=str.lower,
        choices=["song", "artist", "album"],
        help="Specify whether search is for 'song', 'artist' or 'album'.",
    )
    parser.add_argument("terms", type=str, nargs="+", help="Provide terms for search")
    parser.add_argument(
        "--save",
        type=str,
        nargs="*",
        default="json",
        const=None,
        choices=["json", "txt"],
        help="Specify one or more formats to save output: 'json' (default) or 'txt'",
    )
    parser.add_argument(
        "--stdout",
        type=str,
        nargs="?",
        const="json",
        choices=["json", "txt"],
        help="Print output to stdout in 'json' or 'txt' format",
    )
    parser.add_argument(
        "--max-songs",
        type=int,
        help="Specify number of songs when searching for artist",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Turn off the API verbosity"
    )
    args = parser.parse_args()

    # Create an instance of the Genius class
    access_token = os.environ.get("GENIUS_ACCESS_TOKEN", None)
    msg = "Must declare environment variable: GENIUS_ACCESS_TOKEN"
    assert access_token, msg
    api = Genius(access_token, timeout=10)
    if args.quiet:
        api.verbose = False

    # Handle the command-line inputs
    if args.search_type == "song":
        song = api.search_song(*args.terms)
        if not song:
            if not args.quiet:
                print("Could not find specified song. Check spelling?")
            return
        if args.stdout:
            print(song.to_text() if args.stdout == "txt" else song.to_json())
        if args.save:
            for format in args.save:
                if not args.quiet:
                    print(
                        f"Saving lyrics to '{safe_unicode(song.title)}' in {format.upper()} format..."
                    )
                song.save_lyrics(extension=format)
    elif args.search_type == "artist":
        artist = api.search_artist(
            args.terms[0], max_songs=args.max_songs, sort="popularity"
        )
        if not artist:
            if not args.quiet:
                print("Could not find specified artist. Check spelling?")
            return
        if args.stdout:
            print(artist.to_json() if args.stdout == "json" else artist.to_text())
        if args.save:
            for format in args.save:
                if not args.quiet:
                    print(
                        f"Saving '{safe_unicode(artist.name)}' lyrics in {format.upper()} format..."
                    )
                artist.save_lyrics(extension=format)
    elif args.search_type == "album":
        album = api.search_album(*args.terms)
        if not album:
            if not args.quiet:
                print("Could not find specified album. Check spelling?")
            return
        if args.stdout:
            print(album.to_json() if args.stdout == "json" else album.to_text())
        if args.save:
            for format in args.save:
                if not args.quiet:
                    print(
                        f"Saving '{safe_unicode(album.name)}' lyrics in {format.upper()} format..."
                    )
                album.save_lyrics(extension=format)


if __name__ == "__main__":
    main()



class QuestionMethods(object):
    """Question methods of the public API."""

    def questions(
        self,
        album_id=None,
        song_id=None,
        per_page=None,
        page=None,
        state=None,
        text_format=None,
    ):
        """Gets the questions on an album or a song.

        You must supply one of :obj:`album_id` or :obj:`song_id`.

        Args:
            time_period (:obj:`str`, optional): Time period of the results
                ('day', 'week', 'month' or 'all_time').
            per_page (:obj:`int`, optional): Number of results to
                return per request. It can't be more than 50.
            page (:obj:`int`, optional): Paginated offset (number of the page).
            state (:obj:`str`, optional): State of the question.
            text_format (:obj:`str`, optional): Text format of the results
                ('dom', 'html', 'markdown' or 'plain').

        Returns:
            :obj:`dict`

        """
        msg = "Must supply `album_id` or `song_id`."
        assert any([album_id, song_id]), msg
        msg = "Pass only one of `album_id` and `song_id`, not both."
        condition = sum([bool(album_id), bool(song_id)]) == 1
        assert condition, msg
        endpoint = "questions"
        params = {
            "per_page": per_page,
            "page": page,
            "state": state,
            "text_format": text_format or self.response_format,
        }
        if album_id:
            params["album_id"] = album_id
        elif song_id:
            params["song_id"] = song_id
        return self._make_request(path=endpoint, params_=params, public_api=True)
